from __future__ import annotations

import itertools
from typing import Dict, List, Tuple

from numpy import float64

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.coupler import coupler
from gdsfactory.components.crossing_waveguide import compensation_path, crossing45
from gdsfactory.port import get_ports_facing

COUNTER = itertools.count()


def gen_tmp_port_name() -> str:
    return f"{next(COUNTER)}"


def swap(straights, i, j):
    a = straights[i]
    straights[i] = straights[j]
    straights[j] = a
    return straights


def dist(i, wgs1, wgs2):
    a = wgs1.index(i)
    b = wgs2.index(i)
    return b - a


def get_sequence_cross(
    straights_start, straights_end, iter_max: int = 100, symbols=("X", "-")
):
    """Returns sequence of crossings to achieve the permutations between two columns of I/O.

    Args:
        straights_start: list of the input port indices.
        straights_end: list of the output port indices.
        iter_max: maximum iterations.
        symbols: [`X` , `S`].

    Notes:
        symbols to be used in the returned sequence:
        - `X`:represents the crossing symbol: two Xs next to each-other means that the two
            modes have to be swapped
        - `S`:Straight straight, or compensation path typically

    """
    wgs = list(straights_start)
    straights_end = list(straights_end)
    N = len(wgs)
    sequence = []
    X, S = symbols  # Cross, Straight symbols
    nb_iters = 0
    while wgs != straights_end:
        if nb_iters > iter_max:
            print(
                "Exceeded max number of iterations. The following I/O are mismatched:"
            )
            for i in range(len(straights_end)):
                print(wgs[i], "<->", straights_end[i])
            return sequence

        if nb_iters > 2 and sequence[-1] == sequence[-2]:
            print("Two consecutive sequences are the same. Got stuck")
            return sequence

        swaps = []
        i = 0
        # total_dist = 0

        while i < N - 1:
            a = wgs[i]
            b = wgs[i + 1]
            d1 = dist(a, wgs, straights_end)
            d2 = dist(b, wgs, straights_end)
            # total_dist += abs(d1) + abs(d2)

            # The equality cases are very important:
            # if one straight needs to cross, then even if the other one is
            # already at the right place, it must swap to allow the other one
            # to cross

            if d1 >= 0 and d2 <= 0 and (d1 != 0 or d2 != 0):
                wgs = swap(wgs, i, i + 1)
                swaps += [X, X]
                i += 1
            else:
                # Edge case if only one wg remain it is straight
                swaps += [S]
            # We cannot swap twice the same straight on the same iteration, so we
            # skip the next straight by incrementing

            # Edge case: Cannot swap if only one wg left so it has to be a straight
            if i == N - 2:
                swaps += [S]
            i += 1

        sequence.append(swaps)
        nb_iters += 1
    return sequence


def component_sequence_to_str(sequence):
    """Transform a sequence of components (such as the one obtained from.

    get_sequence_cross_str) into an ASCII block which can be used either as
    a cartoon or as an input for component_lattice(lattice = ...)
    """
    component_txt_lattice = ""
    M = len(sequence[0])

    for i in range(M):
        j = M - 1 - i
        line = "".join(col[j] for col in sequence) + "\n"
        component_txt_lattice += line

    return component_txt_lattice


def get_sequence_cross_str(straights_start, straights_end, iter_max: int = 100):
    seq = get_sequence_cross(
        straights_start, straights_end, iter_max=iter_max, symbols=["X", "-"]
    )

    return component_sequence_to_str(seq)


@gf.cell
def component_lattice(
    lattice: str = """
        C-X
        CXX
        CXX
        C-X
        """,
    symbol_to_component: Dict[str, Component] = None,
    grid_per_unit: int = 1000,
) -> Component:
    """Return a lattice Component of N inputs and outputs Columns must have.

    components with the same x spacing between input/output ports Lines must
    have components with the same y spacing between input/output ports.

    Args:
        lattice: ASCII map with character.
        symbol_to_component: dict of ASCII character to component.
        grid_per_unit: int.

    Lattice example:

    .. code::

        X-X
        XCX
        XCX
        X-X

    .. plot::
      :include-source:

      import gdsfactory as gf
      from gdsfactory.components.crossing_waveguide import crossing45
      from gdsfactory.components.crossing_waveguide import compensation_path

      symbol_to_component =  {
            "C": gf.routing.fanout2x2(component=gf.components.coupler(), port_spacing=40.0),
            "X": crossing45(port_spacing=40.0),
            "-": compensation_path(crossing45=crossing45(port_spacing=40.0)),
      }
      c = gf.components.component_lattice(symbol_to_component=symbol_to_component)
      c.plot()
    """
    x = crossing45(port_spacing=40)
    symbol_to_component = symbol_to_component or {
        "C": gf.routing.fanout2x2(component=coupler(), port_spacing=40.0),
        "X": x,
        "-": compensation_path(x),
    }

    # Find y spacing and check that all components have same y spacing
    y_spacing = None
    for component in symbol_to_component.values():
        component = gf.get_component(component)
        # component = component.copy()
        # component.auto_rename_ports_orientation()

        for direction in ["W", "E"]:
            ports_dir = get_ports_facing(component.ports, direction)
            ports_dir.sort(key=lambda p: p.y)
            nb_ports = len(ports_dir)
            if nb_ports > 1:
                _y_spacing = (ports_dir[-1].y - ports_dir[0].y) / (nb_ports - 1)
                if y_spacing is None:
                    y_spacing = _y_spacing
                else:
                    assert abs(y_spacing - _y_spacing) < 0.1 / grid_per_unit, (
                        "All component must have the same y port spacing. Got"
                        f" {y_spacing}, {_y_spacing} for {component.name}"
                    )

    a = y_spacing
    columns, columns_to_length = parse_lattice(lattice, symbol_to_component)
    keys = sorted(columns.keys())

    components_to_nb_input_ports = {
        c: len(get_ports_facing(symbol_to_component[c], "W"))
        for c in symbol_to_component.keys()
    }

    component = gf.Component()
    x = 0
    for i in keys:
        col = columns[i]
        j = 0
        L = columns_to_length[i]
        skip = 0  # number of lines to skip depending on the number of ports
        for c in col:
            y = -j * a
            if skip == 1:
                j += skip
                skip = 0
                continue

            if c in symbol_to_component.keys():
                # Compute the number of ports to skip: They will already be
                # connected since they belong to this component

                nb_inputs = components_to_nb_input_ports[c]
                skip = nb_inputs - 1

                ports_cw = symbol_to_component[c].get_ports_list(clockwise=True)
                _cmp = symbol_to_component[c].ref((x, y), port_id=ports_cw[skip].name)

                # _cmp = symbol_to_component[c].ref((x, y), port_id="oW{}".format(skip))
                component.add(_cmp)

                if i == 0:
                    _ports = get_ports_facing(_cmp, "W")
                    for _p in _ports:
                        component.add_port(gen_tmp_port_name(), port=_p)

                if i == keys[-1]:
                    _ports = get_ports_facing(_cmp, "E")
                    for _p in _ports:
                        component.add_port(gen_tmp_port_name(), port=_p)

            else:
                symbols = list(symbol_to_component.keys())
                raise ValueError(
                    f"symbol {c!r} not in symbol_to_component dict {symbols}"
                )

            j += 1
        x += L

    component.auto_rename_ports()
    return component


def parse_lattice(
    lattice: str, symbol_to_component: Dict[str, Component]
) -> Tuple[Dict[int, List[str]], Dict[int, float64]]:
    """Extract each column.

    Args:
        lattice: string describing lattice.
        symbol_to_component: dict of ASCII character to component.
    """
    lines = lattice.replace(" ", "").split("\n")
    columns = {}
    columns_to_length = {}
    for line in lines:
        if len(line) > 0:
            for i, c in enumerate(line):
                if i not in columns.keys():
                    columns[i] = []

                columns[i].append(c)
                if c in symbol_to_component:
                    cmp = symbol_to_component[c]
                    pcw = cmp.get_ports_list(clockwise=True)
                    pccw = cmp.get_ports_list(clockwise=False)

                    # columns_to_length[i] = cmp.ports["oE0"].x - cmp.ports["oW0"].x
                    columns_to_length[i] = (
                        cmp.ports[pccw[0].name].x - cmp.ports[pcw[0].name].x
                    )

    return columns, columns_to_length


if __name__ == "__main__":
    components_dict = {
        "C": gf.routing.fanout2x2(component=gf.components.coupler(), port_spacing=40.0),
        "X": crossing45(port_spacing=40.0),
        "-": compensation_path(crossing45=crossing45(port_spacing=40.0)),
    }
    c = gf.components.component_lattice(symbol_to_component=components_dict)
    # c= gf.routing.fanout2x2(component=gf.components.coupler(), port_spacing=40.0)
    # c= crossing45(port_spacing=40.0)
    # c = compensation_path(crossing45=crossing45(port_spacing=40.0))
    c.pprint_ports()
    c.show(show_ports=True)



"""components for gdsfactory simphony circuit simulation plugin."""
from __future__ import annotations

from .bend_circular import bend_circular
from .bend_euler import bend_euler
from .coupler import coupler
from .coupler_ring import coupler_ring
from .gc import gc1550te
from .mmi1x2 import mmi1x2
from .mmi2x2 import mmi2x2
from .mzi import mzi
from .ring_double import ring_double
from .ring_single import ring_single
from .straight import straight

model_factory = dict(
    bend_circular=bend_circular,
    bend_euler=bend_euler,
    coupler_ring=coupler_ring,
    coupler=coupler,
    mmi1x2=mmi1x2,
    mmi2x2=mmi2x2,
    straight=straight,
    gc1550te=gc1550te,
)

circuit_factory = dict(mzi=mzi, ring_double=ring_double, ring_single=ring_single)


component_names = list(model_factory.keys())
circuit_names = list(circuit_factory.keys())
__all__ = component_names + circuit_names



from __future__ import annotations

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.read import from_yaml, cell_from_yaml_template
from gdsfactory.routing.factories import routing_strategy
import pathlib

SAMPLE_DIR = pathlib.Path(__file__).parent / "all_angle_routing"


@cell
def demo_all_angle_routing() -> Component:
    """Demonstrate all-angle routing."""
    file = SAMPLE_DIR / "aar_basic_01.pic.yml"
    return from_yaml(file)


def get_yaml_pics():
    files = SAMPLE_DIR.glob("*.pic.yml")
    pics = {}
    for file in files:
        name = file.name[: -len(".pic.yml")]
        pic = cell_from_yaml_template(file, name, routing_strategy=routing_strategy)
        pics[name] = pic
    return pics


if __name__ == "__main__":
    from gdsfactory import grid
    from gdsfactory.pdk import get_active_pdk

    # IMPORTANT: always use this gds write flag when using non-manhattan features
    get_active_pdk().gds_write_settings.flatten_invalid_refs = True

    c = demo_all_angle_routing()
    pics = get_yaml_pics()
    # get all the render-able pics: those with "error" in the name intentionally demonstrate errors
    good_pics = [pic for pic_name, pic in pics.items() if "error" not in pic_name]
    c = grid(good_pics)

    c.show(show_ports=True)



"""Component references.

Adapted from PHIDL https://github.com/amccaugh/phidl/ by Adam McCaughan
"""

from __future__ import annotations

import typing
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import gdstk
import numpy as np
from numpy import cos, float64, int64, mod, ndarray, pi, sin

from gdsfactory.component_layout import Polygon, _GeometryHelper, get_polygons
from gdsfactory.port import (
    Port,
    map_ports_layer_to_orientation,
    map_ports_to_orientation_ccw,
    map_ports_to_orientation_cw,
    select_ports,
)
from gdsfactory.snap import snap_to_grid

if typing.TYPE_CHECKING:
    from gdsfactory.component import Component

Number = Union[float64, int64, float, int]
Coordinate = Union[Tuple[Number, Number], ndarray, List[Number]]
Coordinates = Union[List[Coordinate], ndarray, List[Number], Tuple[Number, ...]]


class SizeInfo:
    def __init__(self, bbox: ndarray) -> None:
        """Initialize this object."""
        self.west = bbox[0, 0]
        self.east = bbox[1, 0]
        self.south = bbox[0, 1]
        self.north = bbox[1, 1]

        self.width = self.east - self.west
        self.height = self.north - self.south

        xc = 0.5 * (self.east + self.west)
        yc = 0.5 * (self.north + self.south)

        self.sw = np.array([self.west, self.south])
        self.se = np.array([self.east, self.south])
        self.nw = np.array([self.west, self.north])
        self.ne = np.array([self.east, self.north])

        self.cw = np.array([self.west, yc])
        self.ce = np.array([self.east, yc])
        self.nc = np.array([xc, self.north])
        self.sc = np.array([xc, self.south])
        self.cc = self.center = np.array([xc, yc])

    def get_rect(
        self, padding=0, padding_w=None, padding_e=None, padding_n=None, padding_s=None
    ) -> Tuple[Coordinate, Coordinate, Coordinate, Coordinate]:
        w, e, s, n = self.west, self.east, self.south, self.north

        padding_n = padding if padding_n is None else padding_n
        padding_e = padding if padding_e is None else padding_e
        padding_w = padding if padding_w is None else padding_w
        padding_s = padding if padding_s is None else padding_s

        w = w - padding_w
        e = e + padding_e
        s = s - padding_s
        n = n + padding_n

        return ((w, s), (e, s), (e, n), (w, n))

    @property
    def rect(self) -> Tuple[Coordinate, Coordinate]:
        return self.get_rect()

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"w: {self.west}\ne: {self.east}\ns: {self.south}\nn: {self.north}\n"


def _rotate_points(
    points: Coordinates,
    angle: float = 45.0,
    center: Coordinate = (
        0.0,
        0.0,
    ),
) -> ndarray:
    """Rotates points around a center point.

    accepts single points [1,2] or array-like[N][2], and will return in kind

    Args:
        points: rotate points around center point.
        angle: in degrees.
        center: x, y.
    """
    # First check for common, easy values of angle
    p_arr = np.asarray(points)
    if angle == 0:
        return p_arr

    c0 = np.asarray(center)
    displacement = p_arr - c0
    if angle == 180:
        return c0 - displacement

    if p_arr.ndim == 2:
        perpendicular = displacement[:, ::-1]
    elif p_arr.ndim == 1:
        perpendicular = displacement[::-1]

    # Fall back to trigonometry
    angle = angle * pi / 180
    ca = cos(angle)
    sa = sin(angle)
    sa = np.array((-sa, sa))
    return displacement * ca + perpendicular * sa + c0


class ComponentReference(_GeometryHelper):
    """Pointer to a Component with x, y, rotation, mirror.

    Args:
        component: Component The referenced Component.
        columns: Number of columns in the array.
        rows: Number of rows in the array.
        spacing: Distances between adjacent columns and adjacent rows.
        origin: array-like[2] of int or float
            Position where the cell is inserted.
        rotation : int or float
            Angle of rotation of the reference (in `degrees`).
        magnification : int or float
            Magnification factor for the reference.
        x_reflection : bool
            If True, the reference is reflected parallel to the x direction
            before being rotated.
        name : str (optional)
            A name for the reference (if provided).

    """

    def __init__(
        self,
        component: Component,
        origin: Coordinate = (0, 0),
        rotation: float = 0,
        magnification: float = 1,
        x_reflection: bool = False,
        visual_label: str = "",
        columns: int = 1,
        rows: int = 1,
        spacing=None,
        name: Optional[str] = None,
        v1: Optional[Tuple[float, float]] = None,
        v2: Optional[Tuple[float, float]] = None,
    ) -> None:
        """Initialize the ComponentReference object."""
        self._reference = gdstk.Reference(
            cell=component._cell,
            origin=origin,
            rotation=np.deg2rad(rotation),
            magnification=magnification,
            x_reflection=x_reflection,
            columns=columns,
            rows=rows,
            spacing=spacing,
        )
        if v1 or v2:
            self._reference.repetition = gdstk.Repetition(
                columns=columns, rows=rows, v1=v1, v2=v2
            )

        self._ref_cell = component
        self._owner = None
        self._name = name

        # The ports of a ComponentReference have their own unique id (uid),
        # since two ComponentReferences of the same parent Component can be
        # in different locations and thus do not represent the same port
        self._local_ports = {
            name: port._copy() for name, port in component.ports.items()
        }
        self.visual_label = visual_label
        # self.uid = str(uuid.uuid4())[:8]

    @property
    def v1(self) -> Optional[Tuple[float, float]]:
        return self._reference.repetition.v1

    @property
    def v2(self) -> Optional[Tuple[float, float]]:
        return self._reference.repetition.v2

    @property
    def rows(self) -> int:
        return self._reference.repetition.rows or 1

    @property
    def columns(self) -> int:
        return self._reference.repetition.columns or 1

    @property
    def spacing(self) -> Optional[Tuple[float, float]]:
        return self._reference.repetition.spacing

    @property
    def ref_cell(self):
        return self._ref_cell

    @property
    def parent(self):
        return self._ref_cell

    @property
    def origin(self):
        return self._reference.origin

    @origin.setter
    def origin(self, value):
        self._reference.origin = value

    @property
    def magnification(self) -> float:
        return self._reference.magnification

    @magnification.setter
    def magnification(self, value):
        self._reference.magnification = value

    @property
    def rotation(self) -> float:
        return np.rad2deg(self._reference.rotation)

    @rotation.setter
    def rotation(self, value):
        self._reference.rotation = np.deg2rad(value)

    @property
    def x_reflection(self) -> bool:
        return self._reference.x_reflection

    @x_reflection.setter
    def x_reflection(self, value):
        self._reference.x_reflection = value

    def _set_ref_cell(self, value):
        self._ref_cell = value
        self._reference.cell = value._cell

    @ref_cell.setter
    def ref_cell(self, value):
        self._set_ref_cell(value)

    @parent.setter
    def parent(self, value):
        self._set_ref_cell(value)

    def get_polygons(
        self,
        by_spec: bool = False,
        depth: Optional[int] = None,
        include_paths: bool = True,
        as_array: bool = True,
    ) -> Union[List[Polygon], Dict[Tuple[int, int], List[Polygon]]]:
        """Return the list of polygons created by this reference.

        Args:
            by_spec : bool or tuple
                If True, the return value is a dictionary with the
                polygons of each individual pair (layer, datatype).
                If set to a tuple of (layer, datatype), only polygons
                with that specification are returned.
            depth : integer or None
                If not None, defines from how many reference levels to
                retrieve polygons.  References below this level will result
                in a bounding box.  If `by_spec` is True the key will be the
                name of the referenced cell.
            include_paths: If True, polygonal representation of paths are also included in the result.
            as_array: when as_array=false, return the Polygon objects instead.
                polygon objects have more information (especially when by_spec=False)
                and are faster to retrieve.

        Returns
            out : list of array-like[N][2] or dictionary
                List containing the coordinates of the vertices of each
                polygon, or dictionary with the list of polygons (if
                `by_spec` is True).

        Note:
            Instances of `FlexPath` and `RobustPath` are also included in
            the result by computing their polygonal boundary.
        """
        return get_polygons(
            instance=self,
            by_spec=by_spec,
            depth=depth,
            include_paths=include_paths,
            as_array=as_array,
        )

    def get_labels(self, depth=None, set_transform=True):
        """Return the list of labels created by this reference.

        Args:
            depth : integer or None
                If not None, defines from how many reference levels to
                retrieve labels from.
            set_transform : bool
                If True, labels will include the transformations from
                the reference.

        Returns:
            out : list of `Label`
                List containing the labels in this cell and its references.
        """
        if set_transform:
            return self._reference.get_labels(depth=depth)
        else:
            return self.parent.get_labels(depth=depth)

    def get_bounding_box(self):
        return self._reference.bounding_box()

    @property
    def settings(self):
        return self.parent.settings

    def get_paths(self, depth=None):
        """Return the list of paths created by this reference.

        Args:
            depth : integer or None
                If not None, defines from how many reference levels to
                retrieve paths from.

        Returns:
            list of `FlexPath` or `RobustPath`
                List containing the paths in this cell and its references.
        """
        return self._reference.get_paths(depth=depth)

    def translate(self, dx, dy):
        x0, y0 = self._reference.origin
        self._reference.origin = (x0 + dx, y0 + dy)
        return self

    def area(self, by_spec=False):
        """Calculate total area.

        Args:
            by_spec : bool
                If True, the return value is a dictionary with the areas
                of each individual pair (layer, datatype).

        Returns:
            out : number, dictionary
                Area of this cell.
        """
        return self._reference.area(by_spec=by_spec)

    @property
    def owner(self):
        return self._owner

    @owner.setter
    def owner(self, value):
        if self.owner is None or value is None:
            self._owner = value
        elif value != self._owner:
            raise ValueError(
                f"Cannot reset owner of a reference once it has already been set!"
                f" Reference: {self}. Current owner: {self._owner}. "
                f"Attempting to re-assign to {value!r}"
            )

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value: str):
        if value != self._name:
            if self.owner and value in self.owner.named_references:
                raise ValueError(
                    f"This reference's owner already has a reference with name {value!r}. Please choose another name."
                )
            if self.owner:
                self.owner._named_references.pop(self._name, None)
                self.owner._named_references[value] = self
            self._name = value

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return (
            'ComponentReference (parent Component "%s", ports %s, origin %s, rotation %s,'
            " x_reflection %s)"
            % (
                self.parent.name,
                list(self.ports.keys()),
                self.origin,
                self.rotation,
                self.x_reflection,
            )
        )

    def to_dict(self):
        d = self.parent.to_dict()
        d.update(
            origin=self.origin,
            rotation=self.rotation,
            magnification=self.magnification,
            x_reflection=self.x_reflection,
        )
        return d

    @property
    def bbox(self):
        """Return the bounding box of the ComponentReference.

        it snaps to 3 decimals in um (0.001um = 1nm precision)
        """
        bbox = self.get_bounding_box()
        if bbox is None:
            bbox = ((0, 0), (0, 0))
        return np.round(bbox, 3)

    @classmethod
    def __get_validators__(cls):
        """Get validators."""
        yield cls.validate

    @classmethod
    def validate(cls, v):
        """Check with pydantic ComponentReference valid type."""
        assert isinstance(
            v, ComponentReference
        ), f"TypeError, Got {type(v)}, expecting ComponentReference"
        return v

    def __getitem__(self, key):
        """Access reference ports."""
        if key not in self.ports:
            ports = list(self.ports.keys())
            raise ValueError(f"{key!r} not in {ports}")

        return self.ports[key]

    @property
    def ports(self) -> Dict[str, Port]:
        """This property allows you to access myref.ports, and receive a copy.

        of the ports dict which is correctly rotated and translated.
        """
        for name, port in self.parent.ports.items():
            port = self.parent.ports[name]
            new_center, new_orientation = self._transform_port(
                port.center,
                port.orientation,
                self.origin,
                self.rotation,
                self.x_reflection,
            )
            if name not in self._local_ports:
                self._local_ports[name] = port.copy()
            self._local_ports[name].center = new_center
            self._local_ports[name].orientation = (
                mod(new_orientation, 360) if new_orientation else new_orientation
            )
            self._local_ports[name].parent = self
        # Remove any ports that no longer exist in the reference's parent
        parent_names = self.parent.ports.keys()
        local_names = list(self._local_ports.keys())
        for name in local_names:
            if name not in parent_names:
                self._local_ports.pop(name)
        return self._local_ports

    @property
    def info(self) -> Dict[str, Any]:
        return self.parent.info

    @property
    def metadata_child(self) -> Dict[str, Any]:
        return self.parent.metadata_child

    @property
    def size_info(self) -> SizeInfo:
        return SizeInfo(self.bbox)

    def pprint_ports(self) -> None:
        """Pretty print component ports."""
        ports_list = self.get_ports_list()
        for port in ports_list:
            print(port)

    def _transform_port(
        self,
        point: ndarray,
        orientation: float,
        origin: Coordinate = (0, 0),
        rotation: Optional[int] = None,
        x_reflection: bool = False,
    ) -> Tuple[ndarray, float]:
        """Apply GDS-type transformation to a port (x_ref)."""
        new_point = np.array(point)
        new_orientation = orientation or 0

        if x_reflection:
            new_point[1] = -new_point[1]
            new_orientation = -orientation
        if rotation is not None:
            new_point = _rotate_points(new_point, angle=rotation, center=[0, 0])
            new_orientation += rotation
        if origin is not None:
            new_point = new_point + np.array(origin)

        if orientation is None:
            new_orientation = orientation
        else:
            new_orientation = mod(new_orientation, 360)

        return new_point, new_orientation

    def _transform_point(
        self,
        point: ndarray,
        origin: Coordinate = (0, 0),
        rotation: Optional[int] = None,
        x_reflection: bool = False,
    ) -> ndarray:
        """Apply GDS-type transformation to a point."""
        new_point = np.array(point)

        if x_reflection:
            new_point[1] = -new_point[1]
        if rotation is not None:
            new_point = _rotate_points(new_point, angle=rotation, center=[0, 0])
        if origin is not None:
            new_point = new_point + np.array(origin)

        return new_point

    def move(
        self,
        origin: Union[Port, Coordinate, str] = (0, 0),
        destination: Optional[Union[Port, Coordinate, str]] = None,
        axis: Optional[str] = None,
    ) -> ComponentReference:
        """Move the ComponentReference from origin point to destination.

        Both origin and destination can be 1x2 array-like, Port, or a key
        corresponding to one of the Ports in this device_ref.

        Args:
            origin: Port, port_name or Coordinate.
            destination: Port, port_name or Coordinate.
            axis: for the movement.

        Returns:
            ComponentReference.
        """
        # If only one set of coordinates is defined, make sure it's used to move things
        if destination is None:
            destination = origin
            origin = (0, 0)

        if isinstance(origin, str):
            if origin not in self.ports:
                raise ValueError(f"{origin} not in {self.ports.keys()}")

            origin = self.ports[origin]
            origin = cast(Port, origin)
            o = origin.center
        elif hasattr(origin, "center"):
            origin = cast(Port, origin)
            o = origin.center
        elif np.array(origin).size == 2:
            o = origin
        else:
            raise ValueError(
                f"move(origin={origin})\n"
                f"Invalid origin = {origin!r} needs to be"
                f"a coordinate, port or port name {list(self.ports.keys())}"
            )

        if isinstance(destination, str):
            if destination not in self.ports:
                raise ValueError(f"{destination} not in {self.ports.keys()}")

            destination = self.ports[destination]
            destination = cast(Port, destination)
            d = destination.center
        if hasattr(destination, "center"):
            destination = cast(Port, destination)
            d = destination.center
        elif np.array(destination).size == 2:
            d = destination

        else:
            raise ValueError(
                f"{self.parent.name}.move(destination={destination!r}) \n"
                f"Invalid destination = {destination!r} needs to be"
                f"a coordinate, a port, or a valid port name {list(self.ports.keys())}"
            )

        # Lock one axis if necessary
        if axis == "x":
            d = (d[0], o[1])
        if axis == "y":
            d = (o[0], d[1])

        # This needs to be done in two steps otherwise floating point errors can accrue
        dxdy = np.array(d) - np.array(o)
        self.origin = np.array(self.origin) + dxdy
        self._bb_valid = False
        return self

    def rotate(
        self,
        angle: float = 45,
        center: Union[Coordinate, str, int] = (0.0, 0.0),
    ) -> ComponentReference:
        """Return rotated ComponentReference.

        Args:
            angle: in degrees.
            center: x, y.
        """
        if angle == 0:
            return self
        if isinstance(center, (int, str)):
            center = self.ports[center].center

        if isinstance(center, Port):
            center = center.center
        self.rotation += angle
        self.rotation %= 360
        self.origin = _rotate_points(self.origin, angle, center)
        self._bb_valid = False
        return self

    def mirror_x(
        self, port_name: Optional[str] = None, x0: Optional[Coordinate] = None
    ) -> ComponentReference:
        """Perform horizontal mirror using x0 or port as axis (default, x0=0).

        This is the default for mirror along X=x0 axis
        """
        if port_name is None and x0 is None:
            x0 = -self.x

        if port_name is not None:
            position = self.ports[port_name]
            x0 = position.x
        self.mirror((x0, 1), (x0, 0))
        return self

    def mirror_y(
        self, port_name: Optional[str] = None, y0: Optional[float] = None
    ) -> ComponentReference:
        """Perform vertical mirror using y0 as axis (default, y0=0)."""
        if port_name is None and y0 is None:
            y0 = 0.0

        if port_name is not None:
            position = self.ports[port_name]
            y0 = position.y
        self.mirror((1, y0), (0, y0))
        return self

    def mirror(
        self,
        p1: Coordinate = (0.0, 1.0),
        p2: Coordinate = (0.0, 0.0),
    ) -> ComponentReference:
        """Mirrors.

        Args:
            p1: point 1.
            p2: point 2.
        """
        if isinstance(p1, Port):
            p1 = p1.center
        if isinstance(p2, Port):
            p2 = p2.center
        p1 = np.array(p1)
        p2 = np.array(p2)
        # Translate so reflection axis passes through origin
        self.origin = self.origin - p1

        # Rotate so reflection axis aligns with x-axis
        angle = np.arctan2((p2[1] - p1[1]), (p2[0] - p1[0])) * 180 / pi
        self.origin = _rotate_points(self.origin, angle=-angle, center=(0, 0))
        self.rotation -= angle

        # Reflect across x-axis
        self.x_reflection = not self.x_reflection
        self.origin = (self.origin[0], -1 * self.origin[1])

        self.rotation = -1 * self.rotation

        # Un-rotate and un-translate
        self.origin = _rotate_points(self.origin, angle=angle, center=(0, 0))
        self.rotation += angle
        self.rotation = self.rotation % 360
        self.origin = self.origin + p1

        self._bb_valid = False
        return self

    def connect(
        self,
        port: Union[str, Port],
        destination: Port,
        overlap: float = 0.0,
        preserve_orientation: bool = False,
    ) -> ComponentReference:
        """Return ComponentReference where port connects to a destination.

        Args:
            port: origin (port, or port name) to connect.
            destination: destination port.
            overlap: how deep does the port go inside.
            preserve_orientation: if True, will not rotate the reference to align the port orientations; reference will keep its orientation pre-connection.

        Returns:
            ComponentReference: with correct rotation to connect to destination.
        """
        # port can either be a string with the name, port index, or an actual Port
        if port in self.ports:
            p = self.ports[port]
        elif isinstance(port, Port):
            p = port
        else:
            ports = list(self.ports.keys())
            raise ValueError(
                f"port = {port!r} not in {self.parent.name!r} ports {ports}"
            )

        if (
            destination.orientation is not None
            and p.orientation is not None
            and not preserve_orientation
        ):
            angle = 180 + destination.orientation - p.orientation
            angle = angle % 360
            self.rotate(angle=angle, center=p.center)

        self.move(origin=p, destination=destination)

        if destination.orientation is not None:
            self.move(
                -overlap
                * np.array(
                    [
                        cos(destination.orientation * pi / 180),
                        sin(destination.orientation * pi / 180),
                    ]
                )
            )

        return self

    def get_ports_list(self, **kwargs) -> List[Port]:
        """Return a list of ports.

        Keyword Args:
            layer: port GDS layer.
            prefix: port name prefix.
            orientation: in degrees.
            width: port width.
            layers_excluded: List of layers to exclude.
            port_type: optical, electrical, ...
            clockwise: if True, sort ports clockwise, False: counter-clockwise.
        """
        return list(select_ports(self.ports, **kwargs).values())

    def get_ports_dict(self, **kwargs) -> Dict[str, Port]:
        """Return a dict of ports.

        Keyword Args:
            layer: port GDS layer.
            prefix: port name prefix.
            orientation: in degrees.
            width: port width.
            layers_excluded: List of layers to exclude.
            port_type: optical, electrical, ...
            clockwise: if True, sort ports clockwise, False: counter-clockwise.
        """
        return select_ports(self.ports, **kwargs)

    @property
    def ports_layer(self) -> Dict[str, str]:
        """Return a mapping from layer0_layer1_E0: portName."""
        return map_ports_layer_to_orientation(self.ports)

    def port_by_orientation_cw(self, key: str, **kwargs):
        """Return port by indexing them clockwise."""
        m = map_ports_to_orientation_cw(self.ports, **kwargs)
        if key not in m:
            raise KeyError(f"{key} not in {list(m.keys())}")
        key2 = m[key]
        return self.ports[key2]

    def port_by_orientation_ccw(self, key: str, **kwargs):
        """Return port by indexing them clockwise."""
        m = map_ports_to_orientation_ccw(self.ports, **kwargs)
        if key not in m:
            raise KeyError(f"{key} not in {list(m.keys())}")
        key2 = m[key]
        return self.ports[key2]

    def snap_ports_to_grid(self, nm: int = 1) -> None:
        for port in self.ports.values():
            port.snap_to_grid(nm=nm)

    def get_ports_xsize(self, **kwargs) -> float:
        """Return xdistance from east to west ports.

        Keyword Args:
            kwargs: orientation, port_type, layer.
        """
        ports_cw = self.get_ports_list(clockwise=True, **kwargs)
        ports_ccw = self.get_ports_list(clockwise=False, **kwargs)
        return snap_to_grid(ports_ccw[0].x - ports_cw[0].x)

    def get_ports_ysize(self, **kwargs) -> float:
        """Returns ydistance from east to west ports."""
        ports_cw = self.get_ports_list(clockwise=True, **kwargs)
        ports_ccw = self.get_ports_list(clockwise=False, **kwargs)
        return snap_to_grid(ports_ccw[0].y - ports_cw[0].y)


def test_move():
    import gdsfactory as gf

    c = gf.Component()
    mzi = c.add_ref(gf.components.mzi())
    bend = c.add_ref(gf.components.bend_euler())
    bend.move("o1", mzi.ports["o2"])


def test_get_polygons():
    import gdsfactory as gf

    ref = gf.components.straight()
    p0 = ref.get_polygons(by_spec="WG", as_array=False)
    p1 = ref.get_polygons(by_spec=(1, 0), as_array=True)
    p2 = ref.get_polygons(by_spec=(1, 0), as_array=False)

    p3 = ref.get_polygons(by_spec=True, as_array=True)[(1, 0)]
    p4 = ref.get_polygons(by_spec=True, as_array=False)[(1, 0)]

    assert len(p1) == len(p2) == len(p3) == len(p4) == 1 == len(p0)
    assert p1[0].dtype == p3[0].dtype == float
    assert isinstance(p2[0], Polygon)
    assert isinstance(p4[0], Polygon)


def test_get_polygons_ref():
    import gdsfactory as gf

    ref = gf.components.straight().ref()
    p0 = ref.get_polygons(by_spec="WG", as_array=False)
    p1 = ref.get_polygons(by_spec=(1, 0), as_array=True)
    p2 = ref.get_polygons(by_spec=(1, 0), as_array=False)

    p3 = ref.get_polygons(by_spec=True, as_array=True)[(1, 0)]
    p4 = ref.get_polygons(by_spec=True, as_array=False)[(1, 0)]

    assert len(p1) == len(p2) == len(p3) == len(p4) == 1 == len(p0)
    assert p1[0].dtype == p3[0].dtype == float
    assert isinstance(p2[0], Polygon)
    assert isinstance(p4[0], Polygon)


def test_pads_no_orientation():
    import gdsfactory as gf

    c = gf.Component("pads_no_orientation")
    pt = c << gf.components.pad()
    pb = c << gf.components.pad()
    pb.connect("pad", pt["pad"])


if __name__ == "__main__":
    # test_get_polygons_ref()
    # test_get_polygons()
    import gdsfactory as gf

    c = gf.Component("parent")
    ref = c << gf.components.straight()
    c.add_ports(ref.ports)
    ref.movex(5)
    # assert c.ports['o1'].center[0] == 5, print(c.ports['o1'])
    print(c.ports["o1"].center)
    c.show(show_ports=True)

    # p = ref.get_polygons(by_spec=(1, 0), as_array=False)

    # c = gf.Component("parent")
    # c2 = gf.Component("child")
    # length = 10
    # width = 0.5
    # layer = (1, 0)
    # c2.add_polygon([(0, 0), (length, 0), (length, width), (0, width)], layer=layer)
    # c << c2

    # c = gf.c.dbr()
    # c.show()

    # import gdsfactory as gf

    # c = gf.Component()
    # mzi = c.add_ref(gf.components.mzi())
    # bend = c.add_ref(gf.components.bend_euler())
    # bend.move("o1", mzi.ports["o2"])
    # bend.move("o1", "o2")



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.simulation.simphony.model_from_gdsfactory import (
    GDSFactorySimphonyWrapper,
)


def coupler_fdtd(c=gf.c.coupler, width=0.5, length=20, gap=0.224):
    r"""Coupler based on Lumerical 3D FDTD simulations.

    Args:
        c: Coupler function
        width:0.5
        length: 4
        gap: 0.2

    .. code::

       W1 __             __ E1
            \           /
             \         /
              ========= gap
             /          \
           _/            \_
        W0      length    E0


    """
    if callable(c):
        c = c(width=width, length=length, gap=gap)
    return GDSFactorySimphonyWrapper(component=c)


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np

    wav = np.linspace(1520, 1570, 1024) * 1e-9
    f = 3e8 / wav
    c = gf.c.coupler(length=20, gap=0.224)
    m = coupler_fdtd(c=c)
    s = m.s_parameters(freqs=f)

    plt.plot(wav, np.abs(s[:, 1] ** 2))
    print(m.pins)
    plt.show()



from __future__ import annotations

from typing import Dict, Iterator, Tuple

import matplotlib.pyplot as plt
import numpy as np
import skfem
from matplotlib.animation import FuncAnimation
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.sparse.linalg import splu
from skfem import (
    Basis,
    BilinearForm,
    ElementTriP0,
    ElementTriP1,
    LinearForm,
    asm,
    condense,
    penalize,
    solve,
)
from skfem.helpers import dot, grad
from skfem.visuals.matplotlib import draw, plot


def solve_thermal(
    mesh_filename: str,
    thermal_conductivity: Dict[str, float],
    specific_conductivity: Dict[str, float],
    thermal_diffusivity: Dict[str, float],
    currents: Dict[str, float],
):
    """Thermal simulation.

    Args:
        mesh_filename: Name of the mesh to load.
        thermal_conductivity: thermal conductivity in W/m‧K.
        specific_conductivity: specific conductivity in S/m.
        thermal_diffusivity:
        currents: current flowing through the layer in A.

    Returns:
        nothing yet, WIP.
    """
    mesh = skfem.Mesh.load(mesh_filename)

    @BilinearForm
    def conduction(u, v, w):
        return dot(w["thermal_conductivity"] * u.grad, grad(v))

    @LinearForm
    def unit_load(v, _):
        return v

    basis = Basis(mesh, ElementTriP1())
    joule_heating_rhs = basis.zeros()
    for domain, current in currents.items():  # sum up the sources for the heating
        core_basis = Basis(mesh, basis.elem, elements=mesh.subdomains[domain])
        asm_core_unit_load = asm(unit_load, core_basis)
        core_area = np.sum(asm_core_unit_load)
        joule_heating = (current / core_area) ** 2 / specific_conductivity[domain]
        joule_heating_rhs += joule_heating * asm_core_unit_load

    basis0 = basis.with_element(ElementTriP0())
    thermal_conductivity_p0 = basis0.zeros()
    for domain in thermal_conductivity:
        thermal_conductivity_p0[
            basis0.get_dofs(elements=domain)
        ] = thermal_conductivity[domain]
    thermal_conductivity_p0 *= 1e-12  # 1e-12 -> conversion from 1/m^2 -> 1/um^2

    thermal_conductivity_lhs = asm(
        conduction,
        basis,
        thermal_conductivity=basis0.interpolate(thermal_conductivity_p0),
    )

    temperature = solve(
        *condense(
            thermal_conductivity_lhs,
            joule_heating_rhs,
            D=basis.get_dofs(mesh.boundaries["bottom"]),
        )
    )

    ax = draw(mesh)
    ax.show()

    ax = draw(mesh, boundaries_only=True)
    plot(basis0, thermal_conductivity_p0 * 1e12, ax=ax, colorbar=True)
    ax.figure.set_size_inches(10, 7)
    ax.set_axis_on()
    ax.figure.tight_layout()
    ax.show()

    ax = draw(mesh, boundaries_only=True)
    plot(basis, joule_heating_rhs, ax=ax, colorbar=True, shading="gouraud")
    ax.figure.set_size_inches(10, 7)
    ax.set_axis_on()
    ax.figure.tight_layout()
    ax.show()

    ax = draw(mesh, boundaries_only=True)
    plot(basis, temperature, ax=ax, colorbar=True, shading="gouraud")
    ax.figure.set_size_inches(10, 7)
    ax.set_axis_on()
    ax.figure.tight_layout()
    ax.show()

    print("max temp steady", np.max(temperature))
    print("average team steady", np.mean(temperature))

    thermal_diffusivity_p0 = basis0.zeros()
    for domain in thermal_diffusivity:
        thermal_diffusivity_p0[basis0.get_dofs(elements=domain)] = thermal_diffusivity[
            domain
        ]

    thermal_diffusivity_p0 *= 1e12  # 1e-12 -> conversion from m^2 -> um^2

    @BilinearForm
    def diffusivity_laplace(u, v, w):
        return dot(grad(u) * w["thermal_conductivity"], grad(v))

    @BilinearForm
    def mass(u, v, w):
        return u * v / (w["thermal_diffusivity"] / w["thermal_conductivity"])

    L = asm(
        diffusivity_laplace,
        basis,
        thermal_diffusivity=basis0.interpolate(thermal_diffusivity_p0),
        thermal_conductivity=basis0.interpolate(thermal_conductivity_p0),
    )
    M = asm(
        mass,
        basis,
        thermal_diffusivity=basis0.interpolate(thermal_diffusivity_p0),
        thermal_conductivity=basis0.interpolate(thermal_conductivity_p0),
    )

    dt = 0.1e-6
    theta = 0.5  # Crank–Nicolson
    steps = 200

    L0, M0 = penalize(L, M, D=basis.get_dofs(mesh.boundaries["bottom"]))
    A = M0 + theta * L0 * dt
    B = M0 - (1 - theta) * L0 * dt

    backsolve = splu(A.T).solve  # .T as splu prefers CSC

    def evolve(
        t: float, u: np.ndarray, heating: np.ndarray
    ) -> Iterator[Tuple[float, np.ndarray]]:
        i = 0
        while True:
            t_temperature[i] = t, np.mean(u)
            i += 1
            yield t, u
            t, u = t + dt, backsolve(B @ u + heating * dt)

    ax = draw(mesh, boundaries_only=True)
    ax.set_axis_on()
    ax = plot(mesh, temperature, ax=ax, shading="gouraud")
    title = ax.set_title("t = 0.00")
    field = ax.get_children()[1]  # vertex-based temperature-colour
    fig = ax.get_figure()

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(field, cax=cax)

    def update(event):
        t, u = event
        title.set_text(f"$t$ = {t * 1e6:.2f}us")
        field.set_array(u)

    t_temperature = np.zeros((steps + 2, 2))
    animation = FuncAnimation(
        fig,
        update,
        evolve(0.0, temperature * 0.01, joule_heating_rhs),
        repeat=False,
        interval=30,
        save_count=steps,
    )
    animation.save("heater_up.gif", "imagemagick")
    t_temperature_up = t_temperature

    t_temperature = np.zeros((steps + 2, 2))
    animation = FuncAnimation(
        fig,
        update,
        evolve(0.0, temperature, 0),
        repeat=False,
        interval=30,
        save_count=steps,
    )
    animation.save("heater_down.gif", "imagemagick")
    t_temperature_down = t_temperature

    plt.figure()
    plt.plot(t_temperature[:-1, 0] * 1e6, t_temperature_up[:-1, 1])
    plt.plot(t_temperature[:-1, 0] * 1e6, t_temperature_down[:-1, 1])
    plt.plot(
        t_temperature[:-1, 0] * 1e6, t_temperature[:-1, 1] * 0 + np.mean(temperature)
    )
    plt.xlabel("Time [us]")
    plt.ylabel("Average temperature offset [T]")
    plt.savefig("heater.svg", bbox_inches="tight")
    plt.show()


if __name__ == "__main__":
    import gmsh

    import gdsfactory as gf
    from gdsfactory.simulation.gmsh.mesh2D import mesh2D

    gf.generic_tech.LAYER_STACK.layers["heater"].thickness = 0.13
    gf.generic_tech.LAYER_STACK.layers["heater"].zmin = 2.2
    print(gf.generic_tech.LAYER_STACK.layers.keys())
    # gf.generic_tech.LAYER_STACK.layers["core"].thickness = 2

    heater1 = gf.components.straight_heater_metal(length=50, heater_width=2)
    heater2 = gf.components.straight_heater_metal(length=50, heater_width=2).move(
        [0, -10]
    )

    heaters = gf.Component("heaters")
    heaters << heater1
    # heaters << heater2
    heaters.show()

    geometry = mesh2D(
        heaters,
        ((25, -25), (25, 25)),
        base_resolution=0.4,
        exclude_layers=((1, 10),),
        padding=(10, 10, 1, 1),
        refine_resolution={(1, 0): 0.01, (47, 0): 0.005},
    )

    gmsh.write("mesh.msh")
    gmsh.clear()
    geometry.__exit__()

    solve_thermal(
        mesh_filename="mesh.msh",
        thermal_conductivity={"(47, 0)": 28, "oxide": 1.38, "(1, 0)": 148},
        specific_conductivity={"(47, 0)_0": 2.3e6},
        thermal_diffusivity={
            "(47, 0)": 28 / 598 / 5240,
            "oxide": 1.38 / 709 / 2203,
            "(1, 0)": 148 / 711 / 2330,
        },
        # specific_heat={"(47, 0)_0": 598, 'oxide': 709, '(1, 0)': 711},
        # density={"(47, 0)_0": 5240, 'oxide': 2203, '(1, 0)': 2330},
        currents={"(47, 0)_0": 0.007},
    )



"""FileWatcher based on watchdog. Looks for changes in files with .pic.yml extension."""

from __future__ import annotations

import importlib
import logging
import pathlib
import sys
import time
import traceback
from typing import Callable, Optional

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from gdsfactory.config import cwd
from gdsfactory.pdk import get_active_pdk


class YamlEventHandler(FileSystemEventHandler):
    """Captures pic.yml file change events."""

    def __init__(self, logger=None, path: Optional[str] = None):
        """Initialize the YAML event handler."""
        super().__init__()

        self.logger = logger or logging.root
        pdk = get_active_pdk()
        pdk.register_cells_yaml(dirpath=path, update=True)

    def update_cell(self, src_path, update: bool = False) -> Callable:
        """Parses a YAML file to a cell function and registers into active pdk.

        Args:
            src_path: the path to the file
            update: if True, will update an existing cell function of the same name without raising an error
        Returns:
            The cell function parsed from the yaml file.

        """
        from gdsfactory.cell import CACHE

        pdk = get_active_pdk()
        print(f"Active PDK: {pdk.name}")
        filepath = pathlib.Path(src_path)
        cell_name = filepath.stem.split(".")[0]
        if cell_name in CACHE:
            CACHE.pop(cell_name)
        parser = pdk.circuit_yaml_parser
        function = parser(filepath, name=cell_name)
        try:
            pdk.register_cells_yaml(**{cell_name: function}, update=update)
        except ValueError as e:
            print(e)
        return function

    def on_moved(self, event):
        super().on_moved(event)

        what = "directory" if event.is_directory else "file"
        if what == "file" and event.dest_path.endswith(".pic.yml"):
            self.logger.info("Moved %s: %s", what, event.src_path)
            self.update_cell(event.dest_path)
            self.get_component(event.src_path)

    def on_created(self, event):
        super().on_created(event)

        what = "directory" if event.is_directory else "file"
        if what == "file" and event.src_path.endswith(".pic.yml"):
            self.logger.info("Created %s: %s", what, event.src_path)
            self.update_cell(event.src_path)
            self.get_component(event.src_path)

    def on_deleted(self, event):
        super().on_deleted(event)

        what = "directory" if event.is_directory else "file"

        if what == "file" and event.src_path.endswith(".pic.yml"):
            self.logger.info("Deleted %s: %s", what, event.src_path)
            pdk = get_active_pdk()
            filepath = pathlib.Path(event.src_path)
            cell_name = filepath.stem.split(".")[0]
            pdk.remove_cell(cell_name)

    def on_modified(self, event):
        super().on_modified(event)

        what = "directory" if event.is_directory else "file"
        if (
            what == "file"
            and event.src_path.endswith(".pic.yml")
            or event.src_path.endswith(".py")
        ):
            self.logger.info("Modified %s: %s", what, event.src_path)
            self.get_component(event.src_path)

    def get_component(self, filepath):
        try:
            filepath = pathlib.Path(filepath)
            if filepath.exists():
                if str(filepath).endswith(".pic.yml"):
                    cell_func = self.update_cell(filepath, update=True)
                    c = cell_func()
                    c.show(show_ports=True)
                    # on_yaml_cell_modified.fire(c)
                    return c
                elif str(filepath).endswith(".py"):
                    d = dict(locals(), **globals())
                    d.update(__name__="__main__")
                    exec(filepath.read_text(), d, d)
                else:
                    print("Changed file {filepath} ignored (not .pic.yml or .py)")

        except Exception as e:
            traceback.print_exc(file=sys.stdout)
            print(e)


def watch(path=cwd, pdk=None) -> None:
    path = str(path)
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    if pdk:
        pdk_module = importlib.import_module(pdk)
        pdk_module.PDK.activate()
    event_handler = YamlEventHandler(path=path)
    observer = Observer()
    observer.schedule(event_handler, path, recursive=True)
    observer.start()
    logging.info(f"Observing {path!r}")
    try:
        while True:
            time.sleep(1)
    finally:
        observer.stop()
        observer.join()


if __name__ == "__main__":
    path = sys.argv[1] if len(sys.argv) > 1 else "."
    watch(path)



"""A GDS layer is a tuple of two integers.

You can:

- Load LayerViews from Klayout XML file (.lyp) (recommended)
- Define your layers in a Pydantic BaseModel


"""
from __future__ import annotations

import os
import pathlib
import re
import typing
import xml.etree.ElementTree as ET
from typing import Dict, Optional, Set, Tuple, Union

import numpy as np
from pydantic import BaseModel, Field, validator
from pydantic.color import Color, ColorType

from gdsfactory.config import logger

if typing.TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, DictStrAny, MappingIntStrAny

PathLike = Union[pathlib.Path, str]

Layer = Tuple[int, int]

_klayout_line_styles = {
    "solid": "",
    "dotted": "*.",
    "dashed": "**..**",
    "dash-dotted": "***..**..***",
    "short dashed": "*..*",
    "short dash-dotted": "**.*.*",
    "long dashed": "*****..*****",
    "dash-double-dotted": "***..*.*..**",
}
_klayout_dither_patterns = {
    "solid": "*",
    "hollow": ".",
    "dotted": "*.\n" ".*",
    "coarsely dotted": "*...\n" "....\n" "..*.\n" "....",
    "left-hatched": "*...\n" ".*..\n" "..*.\n" "...*",
    "lightly left-hatched": "*.......\n"
    ".*......\n"
    "..*.....\n"
    "...*....\n"
    "....*...\n"
    ".....*..\n"
    "......*.\n"
    ".......*",
    "strongly left-hatched dense": "**..\n" ".**.\n" "..**\n" "*..*",
    "strongly left-hatched sparse": "**......\n"
    ".**.....\n"
    "..**....\n"
    "...**...\n"
    "....**..\n"
    ".....**.\n"
    "......**\n"
    "*......*",
    "right-hatched": "*...\n" "...*\n" "..*.\n" ".*..",
    "lightly right-hatched": "*.......\n"
    ".......*\n"
    "......*.\n"
    ".....*..\n"
    "....*...\n"
    "...*....\n"
    "..*.....\n"
    ".*......",
    "strongly right-hatched dense": "**..\n" "*..*\n" "..**\n" ".**.",
    "strongly right-hatched sparse": "**......\n"
    "*......*\n"
    "......**\n"
    ".....**.\n"
    "....**..\n"
    "...**...\n"
    "..**....\n"
    ".**.....",
    "cross-hatched": "*...\n" ".*.*\n" "..*.\n" ".*.*",
    "lightly cross-hatched": "*.......\n"
    ".*.....*\n"
    "..*...*.\n"
    "...*.*..\n"
    "....*...\n"
    "...*.*..\n"
    "..*...*.\n"
    ".*.....*",
    "checkerboard 2px": "**..\n" "**..\n" "..**\n" "..**",
    "strongly cross-hatched sparse": "**......\n"
    "***....*\n"
    "..**..**\n"
    "...****.\n"
    "....**..\n"
    "...****.\n"
    "..**..**\n"
    "***....*",
    "heavy checkerboard": "****....\n"
    "****....\n"
    "****....\n"
    "****....\n"
    "....****\n"
    "....****\n"
    "....****\n"
    "....****",
    "hollow bubbles": ".*...*..\n"
    "*.*.....\n"
    ".*...*..\n"
    "....*.*.\n"
    ".*...*..\n"
    "*.*.....\n"
    ".*...*..\n"
    "....*.*.",
    "solid bubbles": ".*...*..\n"
    "***.....\n"
    ".*...*..\n"
    "....***.\n"
    ".*...*..\n"
    "***.....\n"
    ".*...*..\n"
    "....***.",
    "pyramids": ".*......\n"
    "*.*.....\n"
    "****...*\n"
    "........\n"
    "....*...\n"
    "...*.*..\n"
    "..*****.\n"
    "........",
    "turned pyramids": "****...*\n"
    "*.*.....\n"
    ".*......\n"
    "........\n"
    "..*****.\n"
    "...*.*..\n"
    "....*...\n"
    "........",
    "plus": "..*...*.\n"
    "..*.....\n"
    "*****...\n"
    "..*.....\n"
    "..*...*.\n"
    "......*.\n"
    "*...****\n"
    "......*.",
    "minus": "........\n"
    "........\n"
    "*****...\n"
    "........\n"
    "........\n"
    "........\n"
    "*...****\n"
    "........",
    "22.5 degree down": "*......*\n"
    ".**.....\n"
    "...**...\n"
    ".....**.\n"
    "*......*\n"
    ".**.....\n"
    "...**...\n"
    ".....**.",
    "22.5 degree up": "*......*\n"
    ".....**.\n"
    "...**...\n"
    ".**.....\n"
    "*......*\n"
    ".....**.\n"
    "...**...\n"
    ".**.....",
    "67.5 degree down": "*...*...\n"
    ".*...*..\n"
    ".*...*..\n"
    "..*...*.\n"
    "..*...*.\n"
    "...*...*\n"
    "...*...*\n"
    "*...*...",
    "67.5 degree up": "...*...*\n"
    "..*...*.\n"
    "..*...*.\n"
    ".*...*..\n"
    ".*...*..\n"
    "*...*...\n"
    "*...*...\n"
    "...*...*",
    "22.5 degree cross hatched": "*......*\n"
    ".**..**.\n"
    "...**...\n"
    ".**..**.\n"
    "*......*\n"
    ".**..**.\n"
    "...**...\n"
    ".**..**.",
    "zig zag": "..*...*.\n"
    ".*.*.*.*\n"
    "*...*...\n"
    "........\n"
    "..*...*.\n"
    ".*.*.*.*\n"
    "*...*...\n"
    "........",
    "sine": "..***...\n"
    ".*...*..\n"
    "*.....**\n"
    "........\n"
    "..***...\n"
    ".*...*..\n"
    "*.....**\n"
    "........",
    "heavy unordered": "****.*.*\n"
    "**.****.\n"
    "*.**.***\n"
    "*****.*.\n"
    ".**.****\n"
    "**.***.*\n"
    ".****.**\n"
    "*.*.****",
    "light unordered": "....*.*.\n"
    "..*....*\n"
    ".*..*...\n"
    ".....*.*\n"
    "*..*....\n"
    "..*...*.\n"
    "*....*..\n"
    ".*.*....",
    "vertical dense": "*.\n" "*.\n",
    "vertical": ".*..\n" ".*..\n" ".*..\n" ".*..\n",
    "vertical thick": ".**.\n" ".**.\n" ".**.\n" ".**.\n",
    "vertical sparse": "...*....\n" "...*....\n" "...*....\n" "...*....\n",
    "vertical sparse, thick": "...**...\n" "...**...\n" "...**...\n" "...**...\n",
    "horizontal dense": "**\n" "..\n",
    "horizontal": "....\n" "****\n" "....\n" "....\n",
    "horizontal thick": "....\n" "****\n" "****\n" "....\n",
    "horizontal sparse": "........\n"
    "........\n"
    "........\n"
    "********\n"
    "........\n"
    "........\n"
    "........\n"
    "........\n",
    "horizontal sparse, thick": "........\n"
    "........\n"
    "........\n"
    "********\n"
    "********\n"
    "........\n"
    "........\n"
    "........\n",
    "grid dense": "**\n" "*.\n",
    "grid": ".*..\n" "****\n" ".*..\n" ".*..\n",
    "grid thick": ".**.\n" "****\n" "****\n" ".**.\n",
    "grid sparse": "...*....\n"
    "...*....\n"
    "...*....\n"
    "********\n"
    "...*....\n"
    "...*....\n"
    "...*....\n"
    "...*....\n",
    "grid sparse, thick": "...**...\n"
    "...**...\n"
    "...**...\n"
    "********\n"
    "********\n"
    "...**...\n"
    "...**...\n"
    "...**...\n",
}


class HatchPattern(BaseModel):
    """Custom dither pattern. See KLayout documentation for more info.

    Attributes:
        name: Name of the pattern.
        order: Order of pattern.
        custom_pattern: Pattern defining custom shape.
    """

    name: str
    order: Optional[int] = None
    custom_pattern: Optional[str] = None

    class Config:
        """YAML output uses name as the key."""

        fields = {"name": {"exclude": True}}

    @validator("custom_pattern")
    def check_pattern_klayout(cls, pattern: Optional[str], **kwargs) -> Optional[str]:
        if pattern is None:
            return None
        lines = pattern.splitlines()
        if any(len(list(line)) > 32 for line in lines):
            raise ValueError(
                f"Custom pattern {kwargs['values']['name']} has more than 32 characters."
            )
        return pattern

    def to_klayout_xml(self) -> ET.Element:
        if self.custom_pattern is None:
            raise KeyError(
                f"Cannot write custom hatch/dither pattern {self.name} to KLayout xml format because no pattern is present."
            )

        el = ET.Element("custom-dither-pattern")

        subel = ET.SubElement(el, "pattern")
        lines = self.custom_pattern.splitlines()
        if len(lines) == 1:
            subel.text = lines[0]
        else:
            for line in lines:
                ET.SubElement(subel, "line").text = line

        ET.SubElement(el, "order").text = str(self.order)
        ET.SubElement(el, "name").text = self.name
        return el


class LineStyle(BaseModel):
    """Custom line style. See KLayout documentation for more info.

    Attributes:
        name: Name of the line style.
        order: Order of line style.
        custom_style: Line style to use.
    """

    name: str
    order: Optional[int] = None
    custom_style: Optional[str] = None

    class Config:
        """YAML output uses name as the key."""

        fields = {"name": {"exclude": True}}

    @validator("custom_style")
    def check_pattern(cls, pattern: Optional[str], **kwargs) -> Optional[str]:
        if pattern is None:
            return None

        pattern_list = list(pattern)
        valid_chars = all(char in ["*", "."] for char in pattern_list)
        valid_length = len(pattern_list) <= 32
        if (not valid_chars) or (not valid_length):
            raise ValueError(
                f"Custom line pattern {kwargs['values']['name']} must consist of '*' and '.' characters and be no more than 32 characters long."
            )

        return pattern

    def to_klayout_xml(self) -> ET.Element:
        if self.custom_style is None:
            raise KeyError(
                f"Cannot write custom line style {self.name} to KLayout xml format because no pattern is present."
            )

        el = ET.Element("custom-line-pattern")

        ET.SubElement(el, "pattern").text = str(self.custom_style)
        ET.SubElement(el, "order").text = str(self.order)
        ET.SubElement(el, "name").text = self.name
        return el


class LayerView(BaseModel):
    """KLayout layer properties.

    Docstrings adapted from KLayout documentation:
    https://www.klayout.de/lyp_format.html

    Attributes:
        name: Layer name.
        info: Extra info to include in the LayerView.
        layer: GDSII layer.
        layer_in_name: Whether to display the name as 'name layer/datatype' rather than just the layer.
        width: This is the line width of the frame in pixels (or empty for the default which is 1).
        line_style: This is the number of the line style used to draw the shape boundaries.
            An empty string is "solid line". The values are "Ix" for one of the built-in styles
            where "I0" is "solid", "I1" is "dotted" etc.
        hatch_pattern: This is the number of the dither pattern used to fill the shapes.
            The values are "Ix" for one of the built-in pattern where "I0" is "solid" and "I1" is "clear".
        fill_color: Display color of the layer fill.
        frame_color: Display color of the layer frame.
            Accepts Pydantic Color types. See: https://docs.pydantic.dev/usage/types/#color-type for more info.
        fill_brightness: Brightness of the fill.
        frame_brightness: Brightness of the frame.
        animation: This is a value indicating the animation mode.
            0 is "none", 1 is "scrolling", 2 is "blinking" and 3 is "inverse blinking".
            (Only applies to KLayout layer properties)
        xfill: Whether boxes are drawn with a diagonal cross. (Only applies to KLayout layer properties)
        marked: Whether the entry is marked (drawn with small crosses). (Only applies to KLayout layer properties)
        transparent: Whether the entry is transparent.
        visible: Whether the entry is visible.
        valid: Whether the entry is valid. Invalid layers are drawn but you can't select shapes on those layers.
            (Only applies to KLayout layer properties)
        group_members: Add a list of group members to the LayerView.
    """

    name: Optional[str] = None
    info: Optional[str] = None
    layer: Optional[Layer] = None
    layer_in_name: bool = False
    frame_color: Optional[Color] = None
    fill_color: Optional[Color] = None
    frame_brightness: int = 0
    fill_brightness: int = 0
    hatch_pattern: Union[str, HatchPattern] = "solid"
    line_style: Union[str, LineStyle] = "solid"
    valid: bool = True
    visible: bool = True
    transparent: bool = False
    width: Optional[int] = None
    marked: bool = False
    xfill: bool = False
    animation: int = 0
    group_members: Optional[Dict[str, LayerView]] = Field(default_factory=dict)

    class Config:
        """YAML output uses name as the key."""

        fields = {"name": {"exclude": True}}

    def __init__(
        self,
        gds_layer: Optional[int] = None,
        gds_datatype: Optional[int] = None,
        color: Optional[ColorType] = None,
        brightness: Optional[int] = None,
        **data,
    ):
        """Initialize LayerView object."""
        if (gds_layer is not None) and (gds_datatype is not None):
            if "layer" in data and data["layer"] is not None:
                raise KeyError(
                    "Specify either 'layer' or both 'gds_layer' and 'gds_datatype'."
                )
            data["layer"] = (gds_layer, gds_datatype)

        if color is not None:
            if "fill_color" in data or "frame_color" in data:
                raise KeyError(
                    "Specify either a single 'color' or both 'frame_color' and 'fill_color'."
                )
            data["fill_color"] = data["frame_color"] = color
        if brightness is not None:
            if "fill_brightness" in data or "frame_brightness" in data:
                raise KeyError(
                    "Specify either a single 'brightness' or both 'frame_brightness' and 'fill_brightness'."
                )
            data["fill_brightness"] = data["frame_brightness"] = brightness

        super().__init__(**data)

        # Iterate through all items, adding group members as needed
        for name, field in self.__fields__.items():
            default = field.get_default()
            if isinstance(default, LayerView):
                self.group_members[name] = default

    def dict(
        self,
        *,
        include: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        exclude: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        simplify: bool = True,
    ) -> DictStrAny:
        """Generate a dictionary representation of the model, optionally specifying which fields to include or exclude.

        Specify "simplify" to consolidate fill and frame color/brightness if they are the same.
        """
        _dict = super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

        if simplify:
            replace_keys = ["color", "brightness"]
            for key in replace_keys:
                fill = f"fill_{key}"
                frame = f"frame_{key}"

                if (fill in _dict.keys() and frame in _dict.keys()) and (
                    _dict[fill] == _dict[frame]
                ):
                    _dict[key] = _dict.pop(f"frame_{key}")
                    _dict.pop(f"fill_{key}")
        return _dict

    def __str__(self):
        """Returns a formatted view of properties and their values."""
        return "LayerView:\n\t" + "\n\t".join(
            [f"{k}: {v}" for k, v in self.dict().items()]
        )

    def __repr__(self):
        """Returns a formatted view of properties and their values."""
        return self.__str__()

    def get_alpha(self):
        if not self.visible:
            return 0.0
        elif not self.transparent:
            dither_name = getattr(self.hatch_pattern, "name", self.hatch_pattern)
            if dither_name in ["I0", "solid"]:
                return 0.6
            elif dither_name in ["I1", "hollow"]:
                return 0.1
            else:
                return 0.4
        else:
            return 0.3

    def get_color_dict(self) -> Dict[str, str]:
        from gdsfactory.utils.color_utils import ensure_six_digit_hex_color

        if self.fill_color is not None or self.frame_color is not None:
            return {
                "fill_color": ensure_six_digit_hex_color(self.fill_color.as_hex()),
                "frame_color": ensure_six_digit_hex_color(self.frame_color.as_hex()),
            }
        # Colors generated from here: http://phrogz.net/css/distinct-colors.html
        layer_colors = [
            "#3dcc5c",
            "#2b0fff",
            "#cc3d3d",
            "#e5dd45",
            "#7b3dcc",
            "#cc860c",
            "#73ff0f",
            "#2dccb4",
            "#ff0fa3",
            "#0ec2e6",
            "#3d87cc",
            "#e5520e",
        ]
        color = layer_colors[np.mod(self.layer[0], len(layer_colors))]
        return {"fill_color": color, "frame_color": color}

    def _build_klayout_xml_element(
        self, tag: str, name: str, custom_hatch_patterns: dict, custom_line_styles: dict
    ) -> ET.Element:
        """Get XML Element from attributes."""
        from gdsfactory.utils.color_utils import ensure_six_digit_hex_color

        # If hatch pattern name matches a named (built-in) KLayout pattern, use 'I<idx>' notation
        hatch_name = getattr(self.hatch_pattern, "name", self.hatch_pattern)
        if hatch_name in _klayout_dither_patterns:
            dither_pattern = f"I{list(_klayout_dither_patterns).index(hatch_name)}"
        elif hatch_name in custom_hatch_patterns:
            dither_pattern = f"C{list(custom_hatch_patterns).index(hatch_name)}"
        else:
            logger.warning(
                f"Dither pattern {hatch_name} does not correspond to any KLayout built-in or custom pattern! Using 'solid' instead."
            )
            dither_pattern = "solid"

        # If line style name matches a named (built-in) KLayout pattern, use 'I<idx>' notation
        ls_name = getattr(self.line_style, "name", self.line_style)
        if ls_name in _klayout_line_styles:
            line_style = f"I{list(_klayout_line_styles).index(ls_name)}"
        elif ls_name in custom_line_styles:
            line_style = f"C{list(custom_line_styles).index(ls_name)}"
        else:
            logger.warning(
                f"Line style {ls_name} does not correspond to any KLayout built-in or custom pattern! Using 'solid' instead."
            )
            line_style = "solid"

        frame_color = (
            ensure_six_digit_hex_color(self.frame_color.as_hex())
            if isinstance(self.frame_color, Color)
            else self.frame_color
        )
        fill_color = (
            ensure_six_digit_hex_color(self.fill_color.as_hex())
            if isinstance(self.fill_color, Color)
            else self.fill_color
        )
        prop_dict = {
            "frame-color": frame_color,
            "fill-color": fill_color,
            "frame-brightness": self.frame_brightness,
            "fill-brightness": self.fill_brightness,
            "dither-pattern": dither_pattern,
            "line-style": line_style,
            "valid": str(self.valid).lower(),
            "visible": str(self.visible).lower(),
            "transparent": str(self.transparent).lower(),
            "width": self.width,
            "marked": str(self.marked).lower(),
            "xfill": str(self.xfill).lower(),
            "animation": self.animation,
            "name": f"{name} {self.layer[0]}/{self.layer[1]}"
            if self.layer_in_name
            else name,
            "source": f"{self.layer[0]}/{self.layer[1]}@1"
            if self.layer is not None
            else "*/*@*",
        }
        el = ET.Element(tag)
        for key, value in prop_dict.items():
            subel = ET.SubElement(el, key)

            if value is None:
                continue

            if isinstance(value, bool):
                value = str(value).lower()

            subel.text = str(value)
        return el

    def to_klayout_xml(
        self, custom_hatch_patterns: dict, custom_line_styles: dict
    ) -> ET.Element:
        """Return an XML representation of the LayerView."""
        props = self._build_klayout_xml_element(
            "properties",
            name=self.name,
            custom_hatch_patterns=custom_hatch_patterns,
            custom_line_styles=custom_line_styles,
        )
        for member_name, member in self.group_members.items():
            props.append(
                member._build_klayout_xml_element(
                    "group-members",
                    name=member_name,
                    custom_hatch_patterns=custom_hatch_patterns,
                    custom_line_styles=custom_line_styles,
                )
            )
        return props

    @staticmethod
    def _process_name(
        name: str, layer_pattern: Union[str, re.Pattern]
    ) -> Tuple[Optional[str], Optional[bool]]:
        """Strip layer info from name if it exists.

        Args:
            name: XML-formatted name entry.
            layer_pattern: Regex pattern to match layers with.
        """
        from gdsfactory.name import clean_name

        if not name:
            return None, None

        layer_in_name = False
        match = re.search(layer_pattern, name)
        if match:
            name = name[: match.start()].strip()
            layer_in_name = True
        return clean_name(name, remove_dots=True), layer_in_name

    @staticmethod
    def _process_layer(
        layer: str, layer_pattern: Union[str, re.Pattern]
    ) -> Optional[Layer]:
        """Convert .lyp XML layer entry to a Layer.

        Args:
            layer: XML-formatted layer entry.
            layer_pattern: Regex pattern to match layers with.
        """
        match = re.search(layer_pattern, layer)
        if not match:
            raise OSError(f"Could not read layer {layer}!")
        v = match.group().split("/")
        return None if "*" in v else (int(v[0]), int(v[1]))

    @classmethod
    def from_xml_element(
        cls, element: ET.Element, layer_pattern: Union[str, re.Pattern]
    ) -> Optional[LayerView]:
        """Read properties from .lyp XML and generate LayerViews from them.

        Args:
            element: XML Element to iterate over.
            layer_pattern: Regex pattern to match layers with.
        """
        name, layer_in_name = cls._process_name(
            element.find("name").text, layer_pattern
        )
        if name is None:
            return None

        # Translate KLayout index to hatch name
        hatch_pattern = element.find("dither-pattern").text
        if hatch_pattern and re.match(r"I\d+", hatch_pattern):
            hatch_pattern = list(_klayout_dither_patterns.keys())[
                int(hatch_pattern[1:])
            ]

        # Translate KLayout index to line style name
        line_style = element.find("line-style").text
        if line_style and re.match(r"I\d+", line_style):
            line_style = list(_klayout_line_styles.keys())[int(line_style[1:])]

        lv = LayerView(
            name=name,
            layer=cls._process_layer(element.find("source").text, layer_pattern),
            fill_color=element.find("fill-color").text,
            frame_color=element.find("frame-color").text,
            fill_brightness=element.find("fill-brightness").text or 0,
            frame_brightness=element.find("frame-brightness").text or 0,
            hatch_pattern=hatch_pattern or "solid",
            line_style=line_style or "solid",
            valid=element.find("valid").text,
            visible=element.find("visible").text,
            transparent=element.find("transparent").text,
            width=element.find("width").text,
            marked=element.find("marked").text,
            xfill=element.find("xfill").text,
            animation=element.find("animation").text,
            layer_in_name=layer_in_name,
        )

        # Add only if needed, so we can filter by defaults when dumping to yaml
        group_members = {}
        for member in element.iterfind("group-members"):
            member_lv = cls.from_xml_element(member, layer_pattern)
            group_members[member_lv.name] = member_lv

        if group_members != {}:
            lv.group_members = group_members

        return lv


class LayerViews(BaseModel):
    """A container for layer properties for KLayout layer property (.lyp) files.

    Attributes:
        layer_views: Dictionary of LayerViews describing how to display gds layers.
        custom_dither_patterns: Custom dither patterns.
        custom_line_styles: Custom line styles.
        layer_map: Specify a layer_map to get the layer tuple based on the name of the LayerView, rather than the 'layer' argument.
    """

    layer_views: Dict[str, LayerView] = Field(default_factory=dict)
    custom_dither_patterns: Dict[str, HatchPattern] = Field(default_factory=dict)
    custom_line_styles: Dict[str, LineStyle] = Field(default_factory=dict)
    layer_map: Union[Dict[str, Layer], BaseModel] = Field(default_factory=dict)

    def __init__(
        self,
        filepath: Optional[PathLike] = None,
        layer_map: Union[Dict[str, Layer], BaseModel] = None,
        **data,
    ):
        """Initialize LayerViews object.

        Args:
            filepath: can be YAML or LYP.
            layer_map: Optional layermap.
        """
        if filepath is not None:
            filepath = pathlib.Path(filepath)
            if filepath.suffix == ".lyp":
                lvs = LayerViews.from_lyp(filepath=filepath)
                logger.info(
                    f"Importing LayerViews from KLayout layer properties file: {filepath}."
                )
            elif filepath.suffix in [".yaml", ".yml"]:
                lvs = LayerViews.from_yaml(layer_file=filepath)
                logger.info(f"Importing LayerViews from YAML file: {filepath}.")
            else:
                raise ValueError(f"Unable to load LayerViews from {filepath}.")

            data["layer_views"] = lvs.layer_views
            data["custom_line_styles"] = lvs.custom_line_styles
            data["custom_dither_patterns"] = lvs.custom_dither_patterns

        if isinstance(layer_map, BaseModel):
            layer_map = layer_map.dict()

        if layer_map is not None:
            data["layer_map"] = layer_map

        super().__init__(**data)

        for name in self.dict():
            lv = getattr(self, name)
            if isinstance(lv, LayerView):
                #
                if (self.layer_map is not None) and (name in self.layer_map.keys()):
                    lv_dict = lv.dict(exclude={"layer", "name"})
                    lv = LayerView(layer=self.layer_map[name], name=name, **lv_dict)
                self.add_layer_view(name=name, layer_view=lv)

    def add_layer_view(
        self, name: str, layer_view: Optional[LayerView] = None, **kwargs
    ) -> None:
        """Adds a layer to LayerViews.

        Args:
            name: Name of the LayerView.
            layer_view: LayerView to add.
        """
        if name in self.layer_views:
            raise ValueError(
                f"Adding {name!r} already defined {list(self.layer_views.keys())}"
            )
        if layer_view is None:
            layer_view = LayerView(name=name, **kwargs)
        self.layer_views[name] = layer_view

        # If the dither pattern is a CustomDitherPattern, add it to custom_patterns
        dither_pattern = layer_view.hatch_pattern
        if (
            isinstance(dither_pattern, HatchPattern)
            and dither_pattern not in self.custom_dither_patterns.keys()
        ):
            self.custom_dither_patterns[dither_pattern.name] = dither_pattern

        # If hatch_pattern is the name of a custom pattern, replace string with the CustomDitherPattern
        elif (
            isinstance(dither_pattern, str)
            and dither_pattern in self.custom_dither_patterns.keys()
        ):
            layer_view.hatch_pattern = self.custom_dither_patterns[dither_pattern]

        line_style = layer_view.line_style
        if (
            isinstance(line_style, LineStyle)
            and line_style not in self.custom_line_styles.keys()
        ):
            self.custom_line_styles[line_style.name] = line_style
        elif (
            isinstance(line_style, str) and line_style in self.custom_line_styles.keys()
        ):
            layer_view.line_style = self.custom_line_styles[line_style]

    def get_layer_views(self, exclude_groups: bool = False) -> Dict[str, LayerView]:
        """Return all LayerViews.

        Args:
            exclude_groups: Whether to exclude LayerViews that contain other LayerViews.
        """
        layers = {}
        for name, view in self.layer_views.items():
            if view.group_members and not exclude_groups:
                for member_name, member in view.group_members.items():
                    layers[member_name] = member
                continue
            layers[name] = view
        return layers

    def get_layer_view_groups(self) -> Dict[str, LayerView]:
        """Return the LayerViews that contain other LayerViews."""
        return {name: lv for name, lv in self.layer_views.items() if lv.group_members}

    def __str__(self) -> str:
        """Prints the number of LayerView objects in the LayerViews object."""
        lvs = self.get_layer_views()
        groups = self.get_layer_view_groups()
        return (
            f"LayerViews: {len(lvs)} layers ({len(groups)} groups)\n"
            f"\tCustomDitherPatterns: {list(self.custom_dither_patterns.keys())}\n"
            f"\tCustomLineStyles: {list(self.custom_line_styles.keys())}\n"
        )

    def get(self, name: str) -> LayerView:
        """Returns Layer from name.

        Args:
            name: Name of layer.
        """
        if name not in self.layer_views:
            raise ValueError(f"Layer {name!r} not in {list(self.layer_views.keys())}")
        else:
            return self.layer_views[name]

    def __getitem__(self, val: str):
        """Allows accessing to the layer names like ls['gold2'].

        Args:
            val: Layer name to access within the LayerViews.

        Returns:
            self.layers[val]: LayerView in the LayerViews.

        """
        try:
            return self.get_layer_views()[val]
        except Exception as error:
            raise ValueError(
                f"LayerView {val!r} not in LayerViews {list(self.layer_views.keys())}"
            ) from error

    def get_from_tuple(self, layer_tuple: Layer) -> LayerView:
        """Returns LayerView from layer tuple.

        Args:
            layer_tuple: Tuple of (gds_layer, gds_datatype).

        Returns:
            LayerView.
        """
        tuple_to_name = {v.layer: k for k, v in self.get_layer_views().items()}
        if layer_tuple not in tuple_to_name:
            raise ValueError(
                f"LayerView {layer_tuple} not in {list(tuple_to_name.keys())}"
            )

        name = tuple_to_name[layer_tuple]
        return self.get_layer_views()[name]

    def get_layer_tuples(self) -> Set[Layer]:
        """Returns a tuple for each layer."""
        return {layer.layer for layer in self.get_layer_views().values()}

    def clear(self) -> None:
        """Deletes all layers in the LayerViews."""
        self.layer_views = {}

    def preview_layerset(self, size: float = 100.0, spacing: float = 100.0) -> object:
        """Generates a Component with all the layers.

        Args:
            size: square size in um.
            spacing: spacing between each square in um.

        """
        import gdsfactory as gf

        D = gf.Component(name="layerset", with_uuid=True)
        scale = size / 100
        num_layers = len(self.get_layer_views())
        matrix_size = int(np.ceil(np.sqrt(num_layers)))
        sorted_layers = sorted(
            self.get_layer_views().values(), key=lambda x: (x.layer[0], x.layer[1])
        )
        for n, layer in enumerate(sorted_layers):
            layer_tuple = layer.layer
            R = gf.components.rectangle(
                size=(100 * scale, 100 * scale), layer=layer_tuple
            )
            T = gf.components.text(
                text=f"{layer.name}\n{layer_tuple[0]} / {layer_tuple[1]}",
                size=20 * scale,
                position=(50 * scale, -20 * scale),
                justify="center",
                layer=layer_tuple,
            )

            xloc = n % matrix_size
            yloc = int(n // matrix_size)
            D.add_ref(R).movex((100 + spacing) * xloc * scale).movey(
                -(100 + spacing) * yloc * scale
            )
            D.add_ref(T).movex((100 + spacing) * xloc * scale).movey(
                -(100 + spacing) * yloc * scale
            )
        return D

    def to_lyp(
        self, filepath: Union[str, pathlib.Path], overwrite: bool = True
    ) -> None:
        """Write all layer properties to a KLayout .lyp file.

        Args:
            filepath: to write the .lyp file to (appends .lyp extension if not present).
            overwrite: Whether to overwrite an existing file located at the filepath.

        """
        from gdsfactory.utils.xml_utils import make_pretty_xml

        filepath = pathlib.Path(filepath)

        if os.path.exists(filepath) and not overwrite:
            raise OSError("File exists, cannot write.")

        root = ET.Element("layer-properties")

        for lv in self.layer_views.values():
            root.append(
                lv.to_klayout_xml(
                    custom_hatch_patterns=self.custom_dither_patterns,
                    custom_line_styles=self.custom_line_styles,
                )
            )

        for dp in self.custom_dither_patterns.values():
            root.append(dp.to_klayout_xml())

        for ls in self.custom_line_styles.values():
            root.append(ls.to_klayout_xml())

        filepath.write_bytes(make_pretty_xml(root))

    @staticmethod
    def from_lyp(
        filepath: Union[str, pathlib.Path],
        layer_pattern: Optional[Union[str, re.Pattern]] = None,
    ) -> LayerViews:
        r"""Write all layer properties to a KLayout .lyp file.

        Args:
            filepath: to write the .lyp file to (appends .lyp extension if not present).
            layer_pattern: Regex pattern to match layers with. Defaults to r'(\d+|\*)/(\d+|\*)'.
        """
        layer_pattern = re.compile(layer_pattern or r"(\d+|\*)/(\d+|\*)")

        filepath = pathlib.Path(filepath)

        if not os.path.exists(filepath):
            raise OSError("File not found!")

        tree = ET.parse(filepath)
        root = tree.getroot()
        if root.tag != "layer-properties":
            raise OSError("Layer properties file incorrectly formatted, cannot read.")

        dither_patterns = {}
        for dither_block in root.iter("custom-dither-pattern"):
            name = dither_block.find("name").text
            order = dither_block.find("order").text

            if name is None or order is None:
                continue
            pattern = "\n".join(
                [line.text for line in dither_block.find("pattern").iter()]
            )

            if name in dither_patterns:
                logger.warning(
                    f"Dither pattern named {name!r} already exists. Keeping only the first defined."
                )
                continue

            dither_patterns[name] = HatchPattern(
                name=name,
                order=int(order),
                custom_pattern=pattern.lstrip(),
            )
        line_styles = {}
        for line_block in root.iter("custom-line-style"):
            name = line_block.find("name").text
            order = line_block.find("order").text

            if name is None or order is None:
                continue

            if name in line_styles:
                logger.warning(
                    f"Line style named {name!r} already exists. Keeping only the first defined."
                )
                continue

            line_styles[name] = LineStyle(
                name=name,
                order=int(order),
                custom_style=line_block.find("pattern").text,
            )

        layer_views = {}
        for properties_element in root.iter("properties"):
            lv = LayerView.from_xml_element(
                properties_element, layer_pattern=layer_pattern
            )
            if lv:
                layer_views[lv.name] = lv

        return LayerViews(
            layer_views=layer_views,
            custom_dither_patterns=dither_patterns,
            custom_line_styles=line_styles,
        )

    def to_yaml(
        self, layer_file: Union[str, pathlib.Path], prefer_named_color: bool = True
    ) -> None:
        """Export layer properties to two yaml files.

        Args:
            layer_file: Name of the file to write LayerViews to.
            prefer_named_color: Write the name of a color instead of its hex representation when possible.
        """
        import yaml

        from gdsfactory.utils.yaml_utils import (
            add_color_yaml_presenter,
            add_multiline_str_yaml_presenter,
            add_tuple_yaml_presenter,
        )

        lf_path = pathlib.Path(layer_file)

        add_tuple_yaml_presenter()
        add_multiline_str_yaml_presenter()
        add_color_yaml_presenter(prefer_named_color=prefer_named_color)

        lvs = {
            name: lv.dict(exclude_none=True, exclude_defaults=True, exclude_unset=True)
            for name, lv in self.layer_views.items()
        }

        out_dict = {
            "LayerViews": lvs,
            "CustomDitherPatterns": {
                name: dp.dict(
                    exclude_none=True, exclude_defaults=True, exclude_unset=True
                )
                for name, dp in self.custom_dither_patterns.items()
            },
            "CustomLineStyles": {
                name: ls.dict(
                    exclude_none=True, exclude_defaults=True, exclude_unset=True
                )
                for name, ls in self.custom_line_styles.items()
            },
        }

        lf_path.write_bytes(
            yaml.dump(
                out_dict,
                indent=2,
                sort_keys=False,
                default_flow_style=False,
                encoding="utf-8",
            )
        )

    @staticmethod
    def from_yaml(layer_file: Union[str, pathlib.Path]) -> LayerViews:
        """Import layer properties from two yaml files.

        Args:
            layer_file: Name of the file to read LayerViews, CustomDitherPatterns, and CustomLineStyles from.
        """
        from omegaconf import OmegaConf

        layer_file = pathlib.Path(layer_file)

        properties = OmegaConf.to_container(OmegaConf.load(layer_file.open()))
        lvs = {}
        for name, lv in properties["LayerViews"].items():
            if "group_members" in lv:
                lv["group_members"] = {
                    member_name: LayerView(name=member_name, **member_view)
                    for member_name, member_view in lv["group_members"].items()
                }
            lvs[name] = LayerView(name=name, **lv)

        return LayerViews(
            layer_views=lvs,
            custom_dither_patterns={
                name: HatchPattern(name=name, **dp)
                for name, dp in properties["CustomDitherPatterns"].items()
            },
            custom_line_styles={
                name: LineStyle(name=name, **ls)
                for name, ls in properties["CustomLineStyles"].items()
            },
        )


def _name_to_short_name(name_str: str) -> str:
    """Maps the name entry of the lyp element to a name of the layer.

    i.e. the dictionary key used to access it.
    Default format of the lyp name are:

        - key - layer/datatype - description
        - key - description

    """
    from gdsfactory.name import clean_name

    if name_str is None:
        raise OSError(f"layer {name_str} has no name")
    fields = name_str.split("-")
    name = fields[0].split()[0].strip()
    return clean_name(name, remove_dots=True)


def _name_to_description(name_str) -> str:
    """Gets the description of the layer contained in the lyp name field.

    It is not strictly necessary to have a description. If none there, it returns ''.

    Default format of the lyp name are:

        - key - layer/datatype - description
        - key - description

    """
    if name_str is None:
        raise OSError(f"layer {name_str!r} has no name")
    fields = name_str.split()
    return " ".join(fields[1:]) if len(fields) > 1 else ""


def test_load_lyp():
    from gdsfactory.config import layer_path

    lys = LayerViews(layer_path)
    assert len(lys.layer_views) > 10, len(lys.layer_views)
    return lys


if __name__ == "__main__":
    from gdsfactory.config import PATH, layer_path

    LAYER_VIEWS = LayerViews(filepath=layer_path)
    LAYER_VIEWS.to_yaml(PATH.generic_tech / "layer_views.yaml")

    LAYER_VIEWS2 = LayerViews(filepath=PATH.generic_tech / "layer_views.yaml")
    LAYER_VIEWS2.to_lyp(PATH.repo / "extra" / "test_tech" / "layers.lyp")

    assert LAYER_VIEWS == LAYER_VIEWS2
    # import gdsfactory as gf

    # c = gf.components.rectangle(layer=(123, 0))
    # c.plot()

    # print(LAYER_VIEWS)
    # print(LAYER_STACK.get_from_tuple((1, 0)))
    # print(LAYER_STACK.get_layer_to_material())
    # layer = LayerColor(color="gold")
    # print(layer)

    # lys = test_load_lyp()
    c = LAYER_VIEWS.preview_layerset()
    c.show(show_ports=True)
    # print(LAYERS_OPTICAL)
    # print(layer("wgcore"))
    # print(layer("wgclad"))
    # print(layer("padding"))
    # print(layer("TEXT"))
    # print(type(layer("wgcore")))



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.straight import straight
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@gf.cell
def delay_snake_sbend(
    length: float = 100.0,
    length1: float = 0.0,
    length4: float = 0.0,
    radius: float = 5.0,
    waveguide_spacing: float = 5.0,
    bend: ComponentSpec = "bend_euler",
    sbend: ComponentSpec = "bend_s",
    sbend_xsize: float = 100.0,
    cross_section: CrossSectionSpec = "strip",
    **kwargs,
) -> Component:
    r"""Returns compact Snake with sbend in the middle.

    Input port faces west and output port faces east.

    Args:
        length: total length.
        length1: first straight section length in um.
        length3: third straight section length in um.
        radius: u bend radius in um.
        waveguide_spacing: waveguide pitch in um.
        bend: bend spec.
        sbend: sbend spec.
        sbend_size: sbend size.
        cross_section: cross_section spec.
        kwargs: cross_section settings.

    .. code::

                         length1
         <----------------------------
               length2    spacing    |
                _______              |
               |        \            |
               |          \          | bend1 radius
               |            \sbend   |
          bend2|              \      |
               |                \    |
               |                  \__|
               |
               ---------------------->----------->
                   length3              length4

        We adjust length2 and length3
    """
    c = Component()

    bend180_radius = (radius + waveguide_spacing) / 2
    bend = gf.get_component(
        bend, radius=bend180_radius, angle=180, cross_section=cross_section, **kwargs
    )
    sbend = gf.get_component(
        sbend, size=(sbend_xsize, radius), cross_section=cross_section, **kwargs
    )

    b1 = c << bend
    b2 = c << bend
    bs = c << sbend
    bs.mirror()

    length23 = (
        length - (2 * bend.info["length"] - bs.info["length"]) - length1 - length4
    )
    length2 = length23 / 2
    length3 = length23 / 2

    if length2 < 0:
        raise ValueError(
            f"length2 = {length2} < 0. You need to reduce length1 = {length1} "
            f"or length3 = {length3} or increase length = {length}"
        )

    straight1 = straight(length=length1, cross_section=cross_section, **kwargs)
    straight2 = straight(length=length2, cross_section=cross_section, **kwargs)
    straight3 = straight(length=length3, cross_section=cross_section, **kwargs)
    straight4 = straight(length=length4, cross_section=cross_section, **kwargs)

    s1 = c << straight1
    s2 = c << straight2
    s3 = c << straight3
    s4 = c << straight4

    b1.connect("o2", s1.ports["o2"])
    bs.connect("o2", b1.ports["o1"])

    s2.connect("o2", bs.ports["o1"])

    b2.connect("o1", s2.ports["o1"])
    s3.connect("o1", b2.ports["o2"])
    s4.connect("o1", s3.ports["o2"])

    c.add_port("o1", port=s1.ports["o1"])
    c.add_port("o2", port=s4.ports["o2"])

    c.info["min_bend_radius"] = sbend.info["min_bend_radius"]
    c.info["bend180_radius"] = bend180_radius
    return c


def test_delay_snake_sbend_length():
    import numpy as np

    length = 200.0
    c = delay_snake_sbend(length=length, cross_section="strip_no_pins")
    length_computed = c.area() / 0.5
    np.isclose(length, length_computed)
    return c


if __name__ == "__main__":
    c = test_delay_snake_sbend_length()
    # c = gf.grid(
    #     [
    #         delay_snake_sbend(length=length, cross_section="rib")
    #         for length in [500, 3000]
    #     ]
    # )
    c.show(show_ports=True)



"""Small demonstration of the slot cross_section utilizing add_center_section=False."""

from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    wg1 = gf.components.straight(length=10, width=0.8, cross_section="slot")
    wg1.show(show_ports=True)  # show it in klayout



from typing import Optional

import numpy as np

from gdsfactory.pdk import get_layer_stack
from gdsfactory.technology import LayerStack


class Parameter:
    def __init__(
        self,
        min_value: Optional[float] = None,
        max_value: Optional[float] = None,
        nominal_value: Optional[float] = None,
        step: Optional[float] = None,
    ) -> None:
        """Generic parameter class for training Component models.

        Arguments:
            min_value: minimum value of the parameter. Default to layer thickness minus tolerance.
            max_value: maximum value of the parameter. Default to layer thickness plus tolerance.
            nominal_value: nominal value of the parameter. Default to layer thickness.
            step: size of the step going from min_value to max_value when generating data. Default to 3 steps between min and max.
        """

    def sample(self, rand_val: float = None):
        """Return a random value within a parameter allowable values.

        User can provide their own random number between 0 and 1 (mapping to a value between min and max), or default to uniform sampling.

        Arguments:
            rand_val: random value between 0 and 1, where 0 maps to min_value and 1 to max_value.

        """
        rand_val = rand_val or np.random.rand(1)[0]
        return self.min_value + (self.max_value - self.min_value) * rand_val

    def count(self):
        """Given min, max, and step, returns number of grid points."""
        return np.ceil(np.abs(self.max_value - self.min_value) / self.step)

    def arange(self):
        """Given min, max, and step, return array of values between min and max (inclusive)."""
        return np.arange(self.min_value, self.max_value + self.step / 2, self.step)


class LayerStackThickness(Parameter):
    def __init__(
        self,
        layerstack: Optional[LayerStack] = None,
        layername: Optional[str] = "core",
        **kwargs,
    ) -> None:
        """Layerstack thickness parameter.

        Arguments:
            layerstack: LayerStack
            layername: Name of the layer in the layerstack
        """
        super().__init__(**kwargs)
        self.layerstack = layerstack or get_layer_stack()
        self.layername = layername
        self.min_value = (
            kwargs["min_value"]
            or self.layerstack.layers[self.layername].thickness
            - self.layerstack.layers[self.layername].thickness_tolerance
        )
        self.max_value = (
            kwargs["max_value"]
            or self.layerstack.layers[self.layername].thickness
            + self.layerstack.layers[self.layername].thickness_tolerance
        )
        self.nominal_value = self.layerstack.layers[self.layername].thickness
        self.step = kwargs["step"] or np.abs(self.max_value - self.min_value) / 3
        self.current_value = None
        return None


class NamedParameter(Parameter):
    def __init__(self, **kwargs) -> None:
        """Parameter associated with the Component or simulation (e.g. wavelength)."""
        super().__init__(**kwargs)
        self.min_value = kwargs.get("min_value")
        self.max_value = kwargs.get("max_value")
        self.nominal_value = kwargs.get("nominal_value")
        self.step = kwargs.get("step")
        return None



from __future__ import annotations

import pathlib
import tempfile
from functools import lru_cache

import gdstk

from gdsfactory.component import Component, Port
from gdsfactory.read.import_gds import import_gds
from gdsfactory.typings import Layer


@lru_cache(maxsize=None)
def from_gdstk(cell: gdstk.Cell, **kwargs) -> Component:
    """Returns gdsfactory Component from a gdstk cell.

    Args:
        cell: gdstk cell.

    Keyword Args:
        cellname: cell of the name to import (None) imports top cell.
        snap_to_grid_nm: snap to different nm grid (does not snap if False).
        gdsdir: optional GDS directory.
        read_metadata: loads metadata if it exists.
        hashed_name: appends a hash to a shortened component name.
        kwargs: extra to add to component.info (polarization, wavelength ...).
    """
    with tempfile.TemporaryDirectory() as gdsdir:
        gdsdir = pathlib.Path(gdsdir)
        gdsdir.mkdir(exist_ok=True)
        gdspath = gdsdir / f"{cell.name}.gds"
        filepath = cell.write_gds(gdspath)
        return import_gds(filepath, **kwargs)


@lru_cache(maxsize=None)
def from_phidl(component, port_layer: Layer = (1, 0), **kwargs) -> Component:
    """Returns gdsfactory Component from a phidl Device or function.

    Args:
        component: phidl component.
        port_layer: to add to component ports.

    Keyword Args:
        cellname: cell of the name to import (None) imports top cell.
        snap_to_grid_nm: snap to different nm grid (does not snap if False).
        gdsdir: optional GDS directory.
        read_metadata: loads metadata if it exists.
        hashed_name: appends a hash to a shortened component name.
        kwargs: extra to add to component.info (polarization, wavelength ...).
    """
    device = component() if callable(component) else component

    with tempfile.TemporaryDirectory() as gdsdir:
        gdsdir = pathlib.Path(gdsdir)
        gdsdir.mkdir(exist_ok=True)
        gdspath = gdsdir / f"{device.name}.gds"
        filepath = device.write_gds(gdspath, cellname=device.name)

        component = import_gds(filepath, cache=False, **kwargs)
        component.unlock()

    for p in device.ports.values():
        if p.name not in component.ports:
            component.add_port(
                port=Port(
                    name=p.name,
                    center=p.midpoint,
                    width=p.width,
                    orientation=p.orientation,
                    parent=p.parent,
                    layer=port_layer,
                )
            )
    component.lock()
    return component


if __name__ == "__main__":
    import phidl.geometry as pg

    c = pg.rectangle()

    c = pg.snspd()
    c2 = from_phidl(component=c)
    c3 = from_phidl(component=c)
    c2.show(show_ports=True)



"""PCell demo."""

from __future__ import annotations

import gdsfactory as gf


@gf.cell
def mzi_with_bend(radius: float = 10):
    """Returns MZI interferometer with bend."""
    c = gf.Component()
    mzi = c.add_ref(gf.components.mzi())
    bend = c.add_ref(gf.components.bend_euler(radius=radius))
    bend.connect("o1", mzi.ports["o2"])
    c.add_port("o1", port=mzi.ports["o1"])
    c.add_port("o2", port=bend.ports["o2"])
    return c


if __name__ == "__main__":
    c = mzi_with_bend(radius=25)
    # c = gf.routing.add_fiber_single(c)
    c.show(show_ports=True)



from typing import Any, Dict, List, Optional, Tuple

from pydantic import BaseModel, Field

from gdsfactory.technology.layer_views import LayerViews


class LayerLevel(BaseModel):
    """Level for 3D LayerStack.

    Parameters:
        layer: (GDSII Layer number, GDSII datatype).
        thickness: layer thickness in um.
        thickness_tolerance: layer thickness tolerance in um.
        zmin: height position where material starts in um.
        material: material name.
        sidewall_angle: in degrees with respect to normal.
        width_to_z: if sidewall_angle, relative z-position
            (0 --> zmin, 1 --> zmin + thickness).
        z_to_bias: parametrizes shrinking/expansion of the design GDS layer
            when extruding from zmin (0) to zmin + thickness (1).
            Defaults no buffering [[0, 1], [0, 0]].
        info: simulation_info and other types of metadata.
            mesh_order: lower mesh order (1) will have priority over higher
                mesh order (2) in the regions where materials overlap.
            refractive_index: refractive_index
                can be int, complex or function that depends on wavelength (um).
            type: grow, etch, implant, or background.
            mode: octagon, taper, round.
                https://gdsfactory.github.io/klayout_pyxs/DocGrow.html
            into: etch into another layer.
                https://gdsfactory.github.io/klayout_pyxs/DocGrow.html
            doping_concentration: for implants.
            resistivity: for metals.
            bias: in um for the etch.
    """

    layer: Optional[Tuple[int, int]]
    thickness: float
    thickness_tolerance: Optional[float] = None
    zmin: float
    material: Optional[str] = None
    sidewall_angle: float = 0.0
    width_to_z: float = 0.0
    z_to_bias: Optional[Tuple[List[float], List[float]]] = None
    info: Dict[str, Any] = {}


class LayerStack(BaseModel):
    """For simulation and 3D rendering.

    Parameters:
        layers: dict of layer_levels.
    """

    layers: Optional[Dict[str, LayerLevel]] = Field(default_factory=dict)

    def __init__(self, **data: Any):
        """Add LayerLevels automatically for subclassed LayerStacks."""
        super().__init__(**data)

        for field in self.dict():
            val = getattr(self, field)
            if isinstance(val, LayerLevel):
                self.layers[field] = val

    def get_layer_to_thickness(self) -> Dict[Tuple[int, int], float]:
        """Returns layer tuple to thickness (um)."""
        return {
            level.layer: level.thickness
            for level in self.layers.values()
            if level.thickness
        }

    def get_layer_to_zmin(self) -> Dict[Tuple[int, int], float]:
        """Returns layer tuple to z min position (um)."""
        return {
            level.layer: level.zmin for level in self.layers.values() if level.thickness
        }

    def get_layer_to_material(self) -> Dict[Tuple[int, int], str]:
        """Returns layer tuple to material name."""
        return {
            level.layer: level.material
            for level in self.layers.values()
            if level.thickness
        }

    def get_layer_to_sidewall_angle(self) -> Dict[Tuple[int, int], str]:
        """Returns layer tuple to material name."""
        return {
            level.layer: level.sidewall_angle
            for level in self.layers.values()
            if level.thickness
        }

    def get_layer_to_info(self) -> Dict[Tuple[int, int], Dict]:
        """Returns layer tuple to info dict."""
        return {level.layer: level.info for level in self.layers.values()}

    def to_dict(self) -> Dict[str, Dict[str, Any]]:
        return {level_name: dict(level) for level_name, level in self.layers.items()}

    def __getitem__(self, key) -> LayerLevel:
        """Access layer stack elements."""
        if key not in self.layers:
            layers = list(self.layers.keys())
            raise ValueError(f"{key!r} not in {layers}")

        return self.layers[key]

    def get_klayout_3d_script(
        self,
        klayout28: bool = True,
        print_to_console: bool = True,
        layer_views: Optional[LayerViews] = None,
        dbu: Optional[float] = 0.001,
    ) -> str:
        """Prints script for 2.5 view KLayout information.

        You can add this information in your tech.lyt take a look at
        gdsfactory/klayout/tech/tech.lyt
        """
        out = ""
        for layer_name, level in self.layers.items():
            layer = level.layer
            zmin = level.zmin
            zmax = zmin + level.thickness

            if layer is None:
                continue

            if dbu:
                rnd_pl = len(str(dbu).split(".")[-1])
                zmin = round(zmin, rnd_pl)
                zmax = round(zmax, rnd_pl)

            if klayout28:
                txt = (
                    f"z("
                    f"input({layer[0]}, {layer[1]}), "
                    f"zstart: {zmin}, "
                    f"zstop: {zmax}, "
                    f"name: '{layer_name}: {level.material} {layer[0]}/{layer[1]}'"
                )
                if layer_views:
                    txt += ", "
                    props = layer_views.get_from_tuple(layer)
                    if props.color.fill == props.color.frame:
                        txt += f"color: {props.color.fill}"
                    else:
                        txt += (
                            f"fill: {props.color.fill}, " f"frame: {props.color.frame}"
                        )

                txt += ")"

            else:
                txt = f"{layer[0]}/{layer[1]}: {zmin} {zmax}"
            out += f"{txt}\n"

            if print_to_console:
                print(txt)
        return out


if __name__ == "__main__":
    import pathlib

    filepath = pathlib.Path(
        "/home/jmatres/gdslib/sp/temp/write_sparameters_meep_mpi.json"
    )
    ls_json = filepath.read_bytes()
    ls2 = LayerStack.parse_raw(ls_json)



from __future__ import annotations

import numpy as np
from simphony.libraries.sipann import Waveguide

import gdsfactory as gf


def bend_euler(
    radius: float = 10.0,
    width: float = 0.5,
    thickness: float = 0.22,
    angle: float = 90,
    sw_angle: float = 90.0,
    **kwargs,
):
    """Return simphony Model for a bend using a straight.

    FIXME. this is fake bend! need to create a better model

    Args:
        radius: Radius of straight in microns.
        width: Width of the straights in microns
        thickness: Thickness of the straights in microns
        angle: Number of deg of circle that bent straight transverses
        sw_angle: Sidewall angle from horizontal in degrees, ie 90 makes a square. Defaults to 90.
        kwargs: geometrical args that this model ignores

    """
    c = gf.c.bend_euler(radius=radius, angle=angle)
    length = c.info["length"] * 1e-6
    angle = np.deg2rad(angle)
    width *= 1e-6
    thickness *= 1e-6

    model = Waveguide(
        width=width, thickness=thickness, sw_angle=sw_angle, length=length
    )
    model.rename_pins("o1", "o2")
    return model


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    from gdsfactory.simulation.simphony.plot_model import plot_model

    c = bend_euler()
    wavelengths = np.linspace(1.5, 1.6) * 1e-6
    plot_model(c, wavelengths=wavelengths)
    plt.show()



from __future__ import annotations

from itertools import combinations
from typing import Dict, Optional, Tuple

import numpy as np
from devsim import (
    add_gmsh_contact,
    add_gmsh_interface,
    add_gmsh_region,
    create_device,
    create_gmsh_mesh,
    finalize_mesh,
    get_node_model_values,
    node_solution,
    set_node_values,
    write_devices,
)
from scipy.interpolate import NearestNDInterpolator

from gdsfactory import Component
from gdsfactory.simulation.devsim.doping import get_doping_info_generic
from gdsfactory.simulation.gmsh import (
    fuse_polygons,
    get_u_bounds_polygons,
    uz_xsection_mesh,
)
from gdsfactory.technology import LayerLevel, LayerStack

um_to_cm = 1e-4


def create_2Duz_simulation(
    component: Component,
    xsection_bounds: Tuple[Tuple[float, float], Tuple[float, float]],
    full_layerstack: LayerStack,
    physical_layerstack: LayerStack,
    doping_info,  # Dict[str, DopingLayerLevel],
    contact_info,
    resolutions: Optional[Dict[str, Dict]] = None,
    mesh_scaling_factor: float = um_to_cm,
    default_resolution_min: float = 0.001,
    default_resolution_max: float = 0.2,
    background_tag: Optional[str] = None,
    temp_file_name: str = "temp.msh2",
    devsim_mesh_name: str = "temp",
    devsim_device_name: str = "temp",
    devsim_simulation_filename: str = "devsim.dat",
    global_meshsize_array: Optional[np.array] = None,
    global_meshsize_interpolant_func: Optional[callable] = NearestNDInterpolator,
):
    # Replace relevant physical entities by contacts
    simulation_layertack = physical_layerstack
    for contact_name, contact_dict in contact_info.items():
        contact_layer = full_layerstack.layers[
            contact_dict["physical_layerlevel_to_replace"]
        ]
        layerlevel = LayerLevel(
            layer=contact_dict["gds_layer"],
            thickness=contact_layer.thickness,
            zmin=contact_layer.zmin,
            material=contact_layer.material,
            sidewall_angle=contact_layer.sidewall_angle,
            info=contact_layer.info,
        )
        simulation_layertack.layers[contact_name] = layerlevel

    # Get structural mesh
    mesh = uz_xsection_mesh(
        component,
        xsection_bounds,
        simulation_layertack,
        resolutions=resolutions,
        mesh_scaling_factor=mesh_scaling_factor,
        default_resolution_min=default_resolution_min,
        default_resolution_max=default_resolution_max,
        background_tag=background_tag,
        filename=temp_file_name,
        global_meshsize_array=global_meshsize_array,
        global_meshsize_interpolant_func=global_meshsize_interpolant_func,
    )

    # Get doping layer bounds
    doping_polygons = {}
    for layername, layer_obj in doping_info.items():
        fused_polygons = fuse_polygons(component, layername, layer_obj.layer)
        bounds = get_u_bounds_polygons(fused_polygons, xsection_bounds)
        doping_polygons[layername] = [np.array(bound) * um_to_cm for bound in bounds]

    # Regions, tagged by material
    regions = {}
    create_gmsh_mesh(file=temp_file_name, mesh=devsim_mesh_name)
    simulation_layerstack_dict = simulation_layertack.to_dict()
    for name, values in simulation_layerstack_dict.items():
        add_gmsh_region(
            mesh=devsim_mesh_name,
            gmsh_name=name,
            region=name,
            material=values["material"],
        )
        if values["material"] in regions:
            regions[values["material"]].append(name)
        else:
            regions[values["material"]] = [name]
    if background_tag:
        simulation_layerstack_dict[background_tag] = {"material": background_tag}
        add_gmsh_region(
            mesh=devsim_mesh_name,
            gmsh_name=background_tag,
            region=background_tag,
            material=background_tag,
        )
        regions[background_tag] = [background_tag]

    # Contacts
    contacts = {}
    for contact_name, contact in contact_info.items():
        layer1 = contact_name
        layer2 = contact["physical_layerlevel_to_contact"]
        interface = f"{layer1}___{layer2}"
        if interface not in mesh.cell_sets_dict.keys():
            interface = f"{layer2}___{layer1}"
        add_gmsh_contact(
            gmsh_name=interface,
            material=simulation_layertack.layers[contact_name].material,
            mesh=devsim_mesh_name,
            name=contact_name,
            region=layer2,
        )
        contacts[contact_name] = interface

    # Interfaces (that are not contacts), labeled by material-material
    interfaces = {}
    for (name1, values1), (name2, values2) in combinations(
        simulation_layerstack_dict.items(), 2
    ):
        interface = f"{name1}___{name2}"
        if interface not in mesh.cell_sets_dict.keys():
            interface = f"{name2}___{name1}"
            if interface not in mesh.cell_sets_dict.keys():
                continue
        # if interface not in contacts.values():
        add_gmsh_interface(
            gmsh_name=interface,
            mesh=devsim_mesh_name,
            name=interface,
            region0=name1,
            region1=name2,
        )
        if interface in interfaces:
            interfaces[(values1["material"], values2["material"])].append(interface)
        else:
            interfaces[(values1["material"], values2["material"])] = [interface]
    finalize_mesh(mesh=devsim_mesh_name)
    create_device(mesh=devsim_mesh_name, device=devsim_device_name)

    # Assign doping fields silicon
    for region_name in regions["si"]:
        xpos = get_node_model_values(
            device=devsim_device_name, region=region_name, name="x"
        )
        # ypos = get_node_model_values(device=devsim_device_name, region="si", name="y")
        acceptor = np.zeros_like(xpos)
        donor = np.zeros_like(xpos)
        for layername, bounds in doping_polygons.items():
            print(layername, bounds)
            for bound in bounds:
                print(bound)
                node_inds = np.intersect1d(
                    np.where(xpos >= bound[0] * np.ones_like(xpos)),
                    np.where(xpos <= bound[1] * np.ones_like(xpos)),
                )
                # Assume step doping for now (does not depend on y, or z)
                if doping_info[layername].type == "Acceptor":
                    acceptor[node_inds] += doping_info[layername].z_profile(0)
                elif doping_info[layername].type == "Donor":
                    donor[node_inds] += doping_info[layername].z_profile(0)
                else:
                    raise ValueError(
                        f'Doping type "{doping_info[layername].type}" not supported.'
                    )
        net_doping = donor - acceptor

        node_solution(device=devsim_device_name, region=region_name, name="Acceptors")
        set_node_values(
            device=devsim_device_name,
            region=region_name,
            name="Acceptors",
            values=acceptor,
        )
        node_solution(device=devsim_device_name, region=region_name, name="Donors")
        set_node_values(
            device=devsim_device_name, region=region_name, name="Donors", values=donor
        )
        node_solution(device=devsim_device_name, region=region_name, name="NetDoping")
        set_node_values(
            device=devsim_device_name,
            region=region_name,
            name="NetDoping",
            values=net_doping,
        )

    # Assign doping field contact
    for contact_name in contacts:
        xpos = get_node_model_values(
            device=devsim_device_name, region=contact_name, name="x"
        )
        node_solution(device=devsim_device_name, region=contact_name, name="NetDoping")
        set_node_values(
            device=devsim_device_name,
            region=contact_name,
            name="NetDoping",
            values=[0] * len(xpos),
        )

    # Cast to cm
    # for region_name in get_region_list(device=devsim_device_name):
    #     xpos = get_node_model_values(
    #         device=devsim_device_name, region=region_name, name="x"
    #     )
    #     set_node_values(
    #         device=devsim_device_name, region=region_name, name="x", values=np.array(xpos)*um_to_cm
    #     )
    #     ypos = get_node_model_values(
    #         device=devsim_device_name, region=region_name, name="y"
    #     )
    #     set_node_values(
    #         device=devsim_device_name, region=region_name, name="y", values=np.array(ypos)*um_to_cm
    #     )

    if devsim_simulation_filename:
        write_devices(file=devsim_simulation_filename, type="tecplot")

    return devsim_device_name, regions, interfaces


if __name__ == "__main__":
    import gdsfactory as gf
    from gdsfactory.generic_tech import get_layer_stack_generic

    # We choose a representative subdomain of the component
    waveguide = gf.Component()
    waveguide.add_ref(
        gf.geometry.trim(
            component=gf.components.straight_pn(length=10, taper=None),
            domain=[[3, -4], [3, 4], [5, 4], [5, -4]],
        )
    )

    waveguide.show()

    # We will restrict the physical mesh to a subset of layers:
    layermap = gf.generic_tech.LayerMap()
    physical_layerstack = LayerStack(
        layers={
            k: get_layer_stack_generic().layers[k]
            for k in (
                "slab90",
                "core",
            )
        }
    )

    # Boundary conditions such as contacts are also defined with GDS polygons.
    # In DEVSIM, contacts must be N-1D, where N is the simulation dimensionality.
    # While complete dummy layers + stack definitions could be used, often we are interested in using existing layers as contact (for instance vias).
    # Since this means different contacts could correspond to the same physical layers, this requires more care in the meshing.

    # Simulation layers for contacts
    anode_layer = (105, 0)
    cathode_layer = (106, 0)

    # We will generate contacts where the vias intersect heavy doping
    via_contact_locations = waveguide.extract(layermap.VIAC)
    NPP_location = waveguide.extract(layermap.NPP)
    PPP_location = waveguide.extract(layermap.PPP)
    waveguide.add_ref(
        gf.geometry.boolean(
            via_contact_locations, NPP_location, operation="AND", layer=anode_layer
        )
    )
    waveguide.add_ref(
        gf.geometry.boolean(
            via_contact_locations, PPP_location, operation="AND", layer=cathode_layer
        )
    )

    # Contact info contains (could be improved)
    # gds_layer to use for this contact
    # physical_layerlevel_1: physical to replace by the contact
    # physical_layerlevel_2: interface to physical_layerlevel_1 to define as the contact
    # two layers defining an interface, and the dummy layer for position refinement
    contact_info = {
        "anode": {
            "gds_layer": anode_layer,
            "physical_layerlevel_to_replace": "via_contact",
            "physical_layerlevel_to_contact": "slab90",
        },
        "cathode": {
            "gds_layer": cathode_layer,
            "physical_layerlevel_to_replace": "via_contact",
            "physical_layerlevel_to_contact": "slab90",
        },
    }

    waveguide.show()

    resolutions = {
        "core": {"resolution": 0.01, "distance": 2},
        "slab90": {"resolution": 0.03, "distance": 1},
    }

    # resolutions["via_contact"] = {"resolution": 0.1, "distance": 1}

    device_name, regions, interfaces = create_2Duz_simulation(
        component=waveguide,
        xsection_bounds=[(4, -4), (4, 4)],
        full_layerstack=get_layer_stack_generic(),
        physical_layerstack=physical_layerstack,
        doping_info=get_doping_info_generic(),
        contact_info=contact_info,
        resolutions=resolutions,
        mesh_scaling_factor=1e-4,
        background_tag="sio2",
    )

    print(device_name)
    print(regions)
    print(interfaces)



import copy
from pathlib import Path
from typing import Dict, Optional, Union

import jax.numpy as jnp
from tqdm.contrib.itertools import product

from gdsfactory.simulation.sax.interpolators import nd_nd_interpolation
from gdsfactory.simulation.sax.parameter import LayerStackThickness, NamedParameter
from gdsfactory.technology import LayerStack
from gdsfactory.typings import PortSymmetries


class Model:
    def __init__(
        self,
        trainable_component: callable,
        layerstack: LayerStack,
        trainable_parameters: Dict[
            str, Union[LayerStackThickness, NamedParameter]
        ] = None,
        non_trainable_parameters: Dict[
            str, Union[LayerStackThickness, NamedParameter]
        ] = None,
        simulation_settings: Optional[Dict[str, Union[float, str, int, Path]]] = None,
        num_modes: int = 2,
        port_symmetries: Optional[PortSymmetries] = None,
    ) -> None:
        """Utility class which simplifies the execution of simulations to build compact models for circuit simulations.

        It contains shared utilities for the different types of Models:
            - Optional simulation hyperparameter tuning routine (TODO)
            - Looping over training variables to generate training examples.
            - Consistent formatting of training examples for model building.

        Other functionality such as
            - Simulation setup, execution, caching/loading
            - Assembly of simulation results into S-parameters
        is solver-dependent, and hence resides in child classes.

        Attributes:
            trainable_component: callable wrapping component associated with model
            layerstack: complete layerstack associated with model
            trainable parameters: parameters that are model features that need to be learned (e.g. width, layer thicknesses, wavelength, materials properties)
            non_trainable_parameters: parameters for input_dict whose effect on S-parameters is known and does not need to be trained (e.g. length for a waveguide)
            simulation_settings: simulation parameters that can impact model quality (e.g. resolution, domain size). Set their converged property to True to skip convergence tuning.
            sim_settings: other simulation settings the solver can use
            num_modes: number of modes to consider for each port in S-parameter calculation.
            port_symmetries: as defined in FDTD solvers. Dict establishing equivalency between S-parameters.
                e.g. {"o1@0,o1@0": ["o2@0,o2@0", "o3@0,o3@0", "o4@0,o4@0"]} means that S22 = S33 = S44 are the same as S11,
                and hence the output vector for model training will only contain the key S11 to reduce the number of variables to fit.
        """
        self.component = trainable_component
        self.layerstack = layerstack
        self.trainable_parameters = trainable_parameters or {}
        self.non_trainable_parameters = non_trainable_parameters or {}
        self.simulation_settings = simulation_settings or {}
        self.num_modes = num_modes

        # Extract inputs and outputs vector size data
        self.size_inputs = len(self.trainable_parameters)
        self.num_ports = len(
            trainable_component(self.get_nominal_dict()).get_ports_list(
                port_type="optical"
            )
        )

        # Extract input and output vector label data
        self.input_vector_labels = list(self.trainable_parameters) + list(
            self.non_trainable_parameters.keys()
        )

        # self.size_outputs = self.num_ports * self.num_modes
        self.port_symmetries = port_symmetries

    def get_nominal_dict(self):
        """Return input_dict of nominal parameter values."""
        return {
            name: parameter.nominal_value
            for name, parameter in self.trainable_parameters.items()
        }

    def get_random_dict(self):
        """Return input_dict of randomly sampled parameter values."""
        return {
            name: parameter.sample()
            for name, parameter in self.trainable_parameters.items()
        }

    def parse_input_dict(self, input_dict):
        """Separates between LayerStackThickness inputs and NamedParameter inputs.

        Args:
            input_dict: key needs to match the keys in self.trainable_parameters
                values are the new values to assign to these parameters.
        """
        param_dict = {}
        layerstack_param_dict = {}
        for key, value in input_dict.items():
            if type(self.trainable_parameters[key]) is NamedParameter:
                param_dict[key] = value
            elif type(self.trainable_parameters[key]) is LayerStackThickness:
                layerstack_param_dict[key] = value
        return param_dict, layerstack_param_dict

    def perturb_layerstack(self, layerstack_param_dict):
        """Returns a temporary LayerStack with a new thickness value for the (currently) LayerStackThickness objects in layerstack_param_dict.

        Args:
            layerstack_param_dict: key needs to match a key in self.trainable_parameters having for value a LayerStackThickness object
                                    value is the thickness to assign to this parameter
        """
        perturbed_layerstack = copy.deepcopy(self.layerstack)
        for key, thickness in layerstack_param_dict.items():
            LayerStackThickness_obj = self.trainable_parameters[key]
            perturbed_layerstack.layers[
                LayerStackThickness_obj.layername
            ].thickness = thickness
        return perturbed_layerstack

    def get_model_input_output(self):
        """Generate training data

        Retrieve the input and output data for training the model by getting results on all trainable parameter input combinations.
        """
        ranges_dict = {
            name: parameter.arange()
            for name, parameter in self.trainable_parameters.items()
        }
        input_vectors = []
        output_vectors = []

        for values in product(*ranges_dict.values(), desc="Getting examples: "):
            # Compute results for this example
            input_dict = dict(
                zip(ranges_dict.keys(), [float(value) for value in values])
            )
            new_inputs, output_vector = self.outputs_from_inputs(input_dict)
            input_vector = list(values)
            input_vector.extend(iter(new_inputs))
            input_vectors.append(input_vector)
            output_vectors.append(output_vector)

        return (
            jnp.array(input_vectors),
            jnp.array(output_vectors),
        )

    def set_nd_nd_interp(self):
        """Returns ND-ND interpolator."""
        input_vectors, output_vectors = self.get_model_input_output()
        self.inference = nd_nd_interpolation(input_vectors, output_vectors)

    def validate(self, num_samples=1):
        """Validates the model by calculating random points within the interpolation range, and comparing to predictions."""
        calculated_outputs = []
        inferred_outputs = []
        validation_inputs = []
        for _index in range(num_samples):
            validation_inputs_local = self.get_random_dict()
            validation_inputs.append(validation_inputs_local)
            calculated_outputs.append(
                self.outputs_from_inputs(validation_inputs_local)[1]
            )
            inferred_outputs_local = [
                self.inference[mode](validation_inputs_local.values())
                for mode in range(self.num_modes)
            ]
            inferred_outputs.append(inferred_outputs_local)

        return validation_inputs, calculated_outputs, inferred_outputs



# from sqlmodel import Field, SQLModel
from gdsfactory.database.models import Base
from sqlalchemy import TIMESTAMP, Column, Integer, text, JSON


class SParameterResults(Base):
    __tablename__ = "s_parameter_results"
    __table_args__ = {"comment": "This table holds scattering parameter results."}

    id = Column(Integer, primary_key=True)
    created = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    updated = Column(
        TIMESTAMP, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )
    n_ports = Column(Integer, nullable=False)
    array = Column(JSON, nullable=False)



from gdsfactory.pdk import _ACTIVE_PDK, get_layer_stack
from gdsfactory.simulation.sax.build_model import Model
from gdsfactory.simulation.gmeep import write_sparameters_meep

_ACTIVE_PDK.materials_index.update(sin=2)


class MeepFDTDModel(Model):
    def __init__(self, **kwargs) -> None:
        """Generic model inferred from MEEP 2.5D FDTD calculations."""
        super().__init__(**kwargs)
        return None

    # def get_Sparameters(self, input_dict, output_vector_labels):
    #     """For FDTD, results are directly S-parameters."""
    #     sp = self.get_results(self, input_dict)
    #     output_vector = [sp[output_key] for output_key in self.output_vector_labels]
    #     output_vectors.append(output_vector)
    #     return sp

    def get_results(self, input_dict):
        """Setup and run a simulation with one set of inputs."""
        param_dict, layerstack_param_dict = self.parse_input_dict(input_dict)
        input_component = self.component(param_dict)
        # input_layerstack = self.perturb_layerstack(layerstack_param_dict)

        wavelengths = self.non_trainable_parameters["wavelength"]

        sim_settings = self.simulation_settings["sim_settings"]
        sim_settings["wavelength_start"] = wavelengths.min_value
        sim_settings["wavelength_stop"] = wavelengths.max_value
        sim_settings["wavelength_points"] = wavelengths.count()

        # TODO: 2.5D setup
        # In this case multiple wavelengths probably cannot be extracted from a single simulation, or need dispersive material
        sp = write_sparameters_meep(input_component, **sim_settings)
        # Reformat with wavelengths as inputs
        new_inputs = []
        output_vectors = []
        for wavelength_index, wavelength in enumerate(sp["wavelengths"]):
            new_inputs.append(wavelength)
            output_vectors.append(
                [
                    sp[output_key][wavelength_index]
                    for output_key in self.output_vector_labels
                ]
            )

        return new_inputs, output_vectors


if __name__ == "__main__":
    import gdsfactory as gf
    from gdsfactory.simulation.sax.parameter import NamedParameter
    from gdsfactory.technology import LayerStack

    c = gf.components.coupler_full(
        coupling_length=0.1, dx=10.0, dy=5.0, gap=0.5, dw=0.0, cross_section="strip"
    )
    c.show(show_ports=True)

    filtered_layerstack = LayerStack(
        layers={
            k: get_layer_stack().layers[k]
            for k in (
                "slab90",
                "core",
                "box",
                "clad",
            )
        }
    )

    def trainable_coupler(parameters):
        return gf.components.coupler_full(
            coupling_length=parameters["coupling_length"],
            gap=parameters["gap"],
            dw=0.0,
        )

    port_symmetries_coupler = {
        "o1@0,o1@0": ["o2@0,o2@0", "o3@0,o3@0", "o4@0,o4@0"],
        "o2@0,o1@0": ["o1@0,o2@0", "o3@0,o4@0", "o4@0,o3@0"],
        "o3@0,o1@0": ["o1@0,o3@0", "o2@0,o4@0", "o4@0,o2@0"],
        "o4@0,o1@0": ["o1@0,o4@0", "o2@0,o3@0", "o3@0,o2@0"],
    }

    sim_settings = dict(
        xmargin=1.0,
        ymargin=1.0,
        is_3d=False,
        port_source_names=["o1"],
        port_symmetries=port_symmetries_coupler,
        run=True,
        overwrite=False,
        layer_stack=filtered_layerstack,
        z=0.1,
    )

    coupler_model = MeepFDTDModel(
        component=trainable_coupler,
        layerstack=filtered_layerstack,
        simulation_settings={
            "sim_settings": sim_settings,
        },
        trainable_parameters={
            "coupling_length": NamedParameter(
                min_value=0.1, max_value=10.1, nominal_value=5, step=10
            ),
            "gap": NamedParameter(
                min_value=0.2, max_value=0.5, nominal_value=0.2, step=0.3
            ),
        },
        non_trainable_parameters={
            "wavelength": NamedParameter(
                min_value=1.54, max_value=1.56, nominal_value=1.55, step=0.01
            ),
        },
        num_modes=1,
    )

    # input_vectors, output_vectors = strip_strip_taper_model.get_data()
    interpolator = coupler_model.set_nd_nd_interp()

    # import jax.numpy as jnp

    # params = jnp.stack(
    #     jnp.broadcast_arrays(
    #         jnp.asarray([6.0, 6.0, 6.0]),
    #         jnp.asarray([1.0, 1.0, 1.0]),
    #         jnp.asarray([1.55, 1.5, 1.55]),
    #         jnp.asarray([0.22, 0.22, 0.19]),
    #     ),
    #     0,
    # )

    # print(interpolator["o1@0,o2@0"](params))



from gdsfactory.config import layer_path
from gdsfactory.generic_tech.layer_map import GenericLayerMap as LayerMap
from gdsfactory.generic_tech.layer_stack import LAYER_STACK
from gdsfactory.technology import LayerViews

LAYER = LayerMap()
LAYER_VIEWS = LayerViews(filepath=layer_path)

PORT_MARKER_LAYER_TO_TYPE = {
    LAYER.PORT: "optical",
    LAYER.PORTE: "dc",
    LAYER.TE: "vertical_te",
    LAYER.TM: "vertical_tm",
}

PORT_LAYER_TO_TYPE = {
    LAYER.WG: "optical",
    LAYER.WGN: "optical",
    LAYER.SLAB150: "optical",
    LAYER.M1: "dc",
    LAYER.M2: "dc",
    LAYER.M3: "dc",
    LAYER.TE: "vertical_te",
    LAYER.TM: "vertical_tm",
}

PORT_TYPE_TO_MARKER_LAYER = {v: k for k, v in PORT_MARKER_LAYER_TO_TYPE.items()}

LAYER_TRANSITIONS = {
    LAYER.WG: "taper",
    LAYER.M3: "taper",
    # (LAYER.)
}


def get_generic_pdk():
    from gdsfactory.components import cells
    from gdsfactory.config import sparameters_path
    from gdsfactory.cross_section import cross_sections
    from gdsfactory.pdk import Pdk, constants

    return Pdk(
        name="generic",
        cells=cells,
        cross_sections=cross_sections,
        layers=LAYER.dict(),
        layer_stack=LAYER_STACK,
        layer_views=LAYER_VIEWS,
        layer_transitions=LAYER_TRANSITIONS,
        sparameters_path=sparameters_path,
        constants=constants,
    )


if __name__ == "__main__":
    pdk = get_generic_pdk()
    print(pdk.name)



from __future__ import annotations

import numpy as np

import gdsfactory as gf
import gdsfactory.simulation.gtidy3d as gt

fiber_port_name = "opt_te_1554_15"


def test_sparameters_grating_coupler(overwrite=True) -> None:
    """Checks Sparameters for a grating_coupler_elliptical_arbitrary in 2D."""
    c = gf.components.grating_coupler_elliptical_arbitrary(
        widths=[0.343] * 25, gaps=[0.345] * 25
    )

    fiber_angle_deg = 20
    offsets = [0]
    dfs = [
        gt.write_sparameters_grating_coupler(
            component=c,
            is_3d=False,
            fiber_angle_deg=fiber_angle_deg,
            fiber_xoffset=fiber_xoffset,
            overwrite=overwrite,
        )
        for fiber_xoffset in offsets
    ]
    sp = dfs[0]

    # Check reasonable reflection/transmission
    assert 1 > np.abs(sp[f"{fiber_port_name}@0,o1@0"]).min() > 0, np.abs(
        sp[f"{fiber_port_name}@0,o1@0"]
    ).min()
    assert (
        0.3 > np.abs(sp[f"{fiber_port_name}@0,{fiber_port_name}@0"]).max() > 0
    ), np.abs(sp[f"{fiber_port_name}@0,{fiber_port_name}@0"]).max()


if __name__ == "__main__":
    overwrite = True
    c = gf.components.grating_coupler_elliptical_arbitrary(
        widths=[0.343] * 25, gaps=[0.345] * 25
    )

    fiber_angle_deg = 20
    offsets = [0, 1]
    dfs = [
        gt.write_sparameters_grating_coupler(
            component=c,
            is_3d=False,
            fiber_angle_deg=fiber_angle_deg,
            fiber_xoffset=fiber_xoffset,
            overwrite=overwrite,
        )
        for fiber_xoffset in offsets
    ]
    sp = dfs[0]

    assert 1 > np.abs(sp[f"{fiber_port_name}@0,o1@0"]).min() > 0, np.abs(
        sp[f"{fiber_port_name}@0,o1@0"]
    ).min()
    assert (
        0.3 > np.abs(sp[f"{fiber_port_name}@0,{fiber_port_name}@0"]).max() > 0
    ), np.abs(sp[f"{fiber_port_name}@0,{fiber_port_name}@0"]).max()



"""When you create components you have to make sure they have unique names.

the cell decorator gives unique names to components that depend on their
parameters

"""

from __future__ import annotations

import gdsfactory as gf


def test_autoname() -> None:
    c1 = gf.components.straight(length=5)
    assert c1.name.split("_")[0] == "straight"



from __future__ import annotations

import numpy as np

import gdsfactory as gf
import gdsfactory.simulation.gtidy3d as gt


def test_sparameters_straight(overwrite=True) -> None:
    """Checks Sparameters for a straight waveguide in 2D."""
    c = gf.components.straight(length=2)
    sp = gt.write_sparameters_1x1(c, overwrite=overwrite, is_3d=False)

    # Check reasonable reflection/transmission
    assert 1 > np.abs(sp["o1@0,o2@0"]).min() > 0.8, np.abs(sp["o1@0,o2@0"]).min()
    assert 0 < np.abs(sp["o1@0,o1@0"]).max() < 0.1, np.abs(sp["o1@0,o1@0"]).max()

    # assert np.allclose(sp["o2@0,o1@0"], 1, atol=1e-02), sp["o2@0,o1@0"]
    # assert np.allclose(sp["o1@0,o1@0"], 0, atol=5e-02), sp["o1@0,o1@0"]
    # assert np.allclose(sp["o2@0,o2@0"], 0, atol=5e-02), sp["o2@0,o2@0"]

    # if dataframe_regression:
    #     dataframe_regression.check(sp)


if __name__ == "__main__":
    # test_sparameters_straight()

    overwrite = True
    c = gf.components.straight(length=3)
    sp = gt.write_sparameters_1x1(c, overwrite=overwrite, is_3d=False)

    # Check reasonable reflection/transmission
    assert 1 > np.abs(sp["o1@0,o2@0"]).min() > 0.8, np.abs(sp["o1@0,o2@0"]).min()
    assert 0 < np.abs(sp["o1@0,o1@0"]).max() < 0.1, np.abs(sp["o1@0,o1@0"]).max()



"""Lets pack a doe and export it with metadata.

by Jan-David Fischbach
email: jan-david.fischbach@blacksemicon.de
date: 26.06.22
"""
from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    doe = gf.components.pack_doe(
        gf.components.mmi1x2,
        settings={"length_taper": [10, 15, 20, 30]},
        function="add_fiber_array",
    )
    doe.show()
    doe.write_gds_with_metadata(f"{__file__[:-3]}/test.gds")



"""gdsfactory interface to simulations."""

from __future__ import annotations

from gdsfactory.simulation import plot
from gdsfactory.simulation.get_effective_indices import get_effective_indices
from gdsfactory.simulation.get_sparameters_path import (
    get_sparameters_data_lumerical,
    get_sparameters_data_meep,
    get_sparameters_data_tidy3d,
    get_sparameters_path_lumerical,
    get_sparameters_path_meep,
    get_sparameters_path_tidy3d,
)

__all__ = [
    "plot",
    "get_sparameters_path_meep",
    "get_sparameters_path_lumerical",
    "get_sparameters_path_tidy3d",
    "get_sparameters_data_meep",
    "get_sparameters_data_lumerical",
    "get_sparameters_data_tidy3d",
    "get_effective_indices",
]



import numpy as np
from psycopg2.extensions import register_adapter, AsIs


def adapt_numpy_float64(numpy_float64):
    return AsIs(numpy_float64)


def adapt_numpy_int64(numpy_int64):
    return AsIs(numpy_int64)


def adapt_numpy_float32(numpy_float32):
    return AsIs(numpy_float32)


def adapt_numpy_int32(numpy_int32):
    return AsIs(numpy_int32)


def adapt_numpy_array(numpy_array):
    return AsIs(tuple(numpy_array))


register_adapter(np.float64, adapt_numpy_float64)
register_adapter(np.int64, adapt_numpy_int64)
register_adapter(np.float32, adapt_numpy_float32)
register_adapter(np.int32, adapt_numpy_int32)
register_adapter(np.ndarray, adapt_numpy_array)



from __future__ import annotations

import asyncio
from pathlib import Path


async def handle_return(
    out_or_err: asyncio.streams.StreamReader,
    log_file: Path = None,
    to_console: bool = True,
) -> None:
    with open(log_file, "w") as f:
        while True:
            # Without this sleep, the program won't exit
            await asyncio.sleep(0)
            data = await out_or_err.readline()
            line = data.decode().strip()
            if line:
                if to_console:
                    print(line)
                f.write(line + "\n")


async def execute_and_stream_output(
    command: str, log_file_dir: Path, log_file_str: str
) -> asyncio.subprocess.Process:
    # Best not to use shell, but I can't get create_subprocess_exec to work here
    proc = await asyncio.create_subprocess_shell(
        command,
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    asyncio.create_task(
        handle_return(proc.stdout, log_file=log_file_dir / f"{log_file_str}_out.log")
    )
    asyncio.create_task(
        handle_return(proc.stderr, log_file=log_file_dir / f"{log_file_str}_err.log")
    )

    # This needs to also handle the "wait_to_finish" flag
    await proc.wait()
    return proc



"""Demo of non-hierarchical circuit simulations."""
from __future__ import annotations

import jax.numpy as jnp
import sax

import gdsfactory as gf


def straight(wl=1.5, length=10.0, neff=2.4) -> sax.SDict:
    """Straight model."""
    return sax.reciprocal({("o1", "o2"): jnp.exp(2j * jnp.pi * neff * length / wl)})


def mmi1x2():
    """Assumes a perfect 1x2 splitter."""
    return sax.reciprocal(
        {
            ("o1", "o2"): 0.5**0.5,
            ("o1", "o3"): 0.5**0.5,
        }
    )


def bend_euler(wl=1.5, length=20.0):
    """Assumes reduced transmission for the euler bend compared to a straight."""
    return {k: 0.99 * v for k, v in straight(wl=wl, length=length).items()}


models = {
    "bend_euler": bend_euler,
    "mmi1x2": mmi1x2,
    "straight": straight,
}


def module(S) -> None:
    for k, v in S.items():
        S[k] = np.abs(v) ** 2


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np

    c = gf.components.mzi()
    c.show(show_ports=True)
    netlist = c.get_netlist()
    circuit, _ = sax.circuit(netlist=netlist, models=models)
    wl = np.linspace(1.5, 1.6)
    S = circuit(wl=wl)

    plt.figure(figsize=(14, 4))
    plt.title("MZI")
    plt.plot(1e3 * wl, jnp.abs(S["o1", "o2"]) ** 2)
    plt.xlabel("λ [nm]")
    plt.ylabel("T")
    plt.grid(True)
    plt.show()



from __future__ import annotations

from typing import Optional, Tuple

from numpy import floor

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.compass import compass
from gdsfactory.components.via import via1, via2, viac
from gdsfactory.typings import ComponentSpec, LayerSpec, LayerSpecs


@gf.cell
def via_stack(
    size: Tuple[float, float] = (11.0, 11.0),
    layers: LayerSpecs = ("M1", "M2", "M3"),
    vias: Optional[Tuple[Optional[ComponentSpec], ...]] = (via1, via2),
    layer_port: LayerSpec = None,
) -> Component:
    """Rectangular via array stack.

    You can use it to connect different metal layers or metals to silicon.
    You can use the naming convention via_stack_layerSource_layerDestination
    contains 4 ports (e1, e2, e3, e4)

    also know as Via array
    http://www.vlsi-expert.com/2017/12/vias.html

    spacing = via.info['spacing']
    enclosure = via.info['enclosure']

    Args:
        size: of the layers.
        layers: layers on which to draw rectangles.
        vias: vias to use to fill the rectangles.
        layer_port: if None assumes port is on the last layer.
    """
    width, height = size
    a = width / 2
    b = height / 2

    layers = layers or []

    if layers:
        layer_port = layer_port or layers[-1]

    c = Component()
    c.height = height
    c.info["size"] = (float(size[0]), float(size[1]))
    c.info["layer"] = layer_port

    for layer in layers:
        if layer == layer_port:
            ref = c << compass(
                size=(width, height), layer=layer, port_type="electrical"
            )
            c.add_ports(ref.ports)
        else:
            ref = c << compass(size=(width, height), layer=layer, port_type="placement")

    vias = vias or []
    for via in vias:
        if via is not None:
            via = gf.get_component(via)

            w, h = via.info["size"]
            g = via.info["enclosure"]
            pitch_x, pitch_y = via.info["spacing"]

            nb_vias_x = (width - w - 2 * g) / pitch_x + 1
            nb_vias_y = (height - h - 2 * g) / pitch_y + 1

            nb_vias_x = int(floor(nb_vias_x)) or 1
            nb_vias_y = int(floor(nb_vias_y)) or 1
            ref = c.add_array(
                via, columns=nb_vias_x, rows=nb_vias_y, spacing=(pitch_x, pitch_y)
            )

            cw = (width - (nb_vias_x - 1) * pitch_x - w) / 2
            ch = (height - (nb_vias_y - 1) * pitch_y - h) / 2
            x0 = -a + cw + w / 2
            y0 = -b + ch + h / 2
            ref.move((x0, y0))

    return c


via_stack_m1_m3 = gf.partial(
    via_stack,
    layers=("M1", "M2", "M3"),
    vias=(via1, via2),
)

via_stack_slab_m3 = gf.partial(
    via_stack,
    layers=("SLAB90", "M1", "M2", "M3"),
    vias=(viac, via1, via2),
)
via_stack_npp_m1 = gf.partial(
    via_stack,
    layers=("WG", "NPP", "M1"),
    vias=(None, None, viac),
)
via_stack_slab_npp_m3 = gf.partial(
    via_stack,
    layers=("SLAB90", "NPP", "M1"),
    vias=(None, None, viac),
)
via_stack_heater_mtop = via_stack_heater_m3 = gf.partial(
    via_stack, layers=("HEATER", "M2", "M3"), vias=(via1, via2)
)


if __name__ == "__main__":
    c = via_stack_m1_m3()
    print(c.to_dict())
    c.show(show_ports=True)
